;; emchat-curl.el --- Download files from log buffer   -*- Emacs-Lisp -*-

;; Copyright (C) 2007 - 2011 Steve Youngs

;; Author:     Steve Youngs <steve@emchat.org>
;; Maintainer: Steve Youngs <steve@emchat.org>
;; Created:    <2007-08-31>
;; Homepage:   http://www.emchat.org/
;; Keywords:   curl download ICQ emchat file

;; This file is part of EMchat.

;; Redistribution and use in source and binary forms, with or without
;; modification, are permitted provided that the following conditions
;; are met:
;;
;; 1. Redistributions of source code must retain the above copyright
;;    notice, this list of conditions and the following disclaimer.
;;
;; 2. Redistributions in binary form must reproduce the above copyright
;;    notice, this list of conditions and the following disclaimer in the
;;    documentation and/or other materials provided with the distribution.
;;
;; 3. Neither the name of the author nor the names of any contributors
;;    may be used to endorse or promote products derived from this
;;    software without specific prior written permission.
;;
;; THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR
;; IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
;; WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
;; DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
;; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
;; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
;; SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
;; BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
;; WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
;; OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
;; IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

;;; Commentary:
;;
;;    Download stuff from URls in your log buffer via ffi-curl.el.
;;    That means you need SXEmacs to use this.

;;; Todo:
;;
;;

;;; Code:
(eval-when-compile
  (require 'emchat-utils)
  (autoload #'curl:download& "ffi-curl")
  (autoload #'curl:easy-getinfo "ffi-curl")
  (defvar curl:errors-alist))

(define-error 'emchat-curl
  (if (featurep 'sxemacs)
      "Your SXEmacs lacks FFI support"
    "FFI not available in your emacs, consider upgrading to SXEmacs"))

(defgroup emchat-curl nil
  "EMchat's FFI/Curl"
  :prefix "emchat-curl-"
  :group 'emchat)

(defcustom emchat-curl-download-directory
  (file-name-as-directory (paths-construct-path
			   '("download") emchat-directory))
  "*Default download directory."
  :type 'directory
  :group 'emchat-curl)

(defcustom emchat-curl-hook-function #'emchat-curl-post-hook
  "This function is called after a successful file download.

It must accept one argument, JOB, which contains the worker-job just
finished."
  :type 'function
  :group 'emchat-curl)

;;; Internal variables

(defvar emchat-curl-file nil)
(defvar emchat-curl-dir nil)

(defun emchat-curl-post-hook (job)
  "Default function called after downloading files."
  (let* ((ctx (get job 'ctx))
	 (resp (curl:easy-getinfo ctx :response-code))
	 (url (curl:easy-getinfo ctx :effective-url))
	 (speed (curl:easy-getinfo ctx :speed-download))
	 (size (curl:easy-getinfo ctx :size-download))
	 (file (expand-file-name emchat-curl-file emchat-curl-dir))
	 (resp-str (cdr (assq resp curl:errors-alist))))
    (if resp-str
	(emchat-log-info "Job failed: %s\nReason: %s" job resp-str)
      (emchat-log-info
       "Job: %s
Detail: #<job :url <%s>
	      :local-file \"%s\"
	      :size %d bytes%s
	      :speed %s>
"
       job url file size
       (cond
	((>= size (* 1024 1024 1024))
	 (format " (%0.2fGB)" (/ size (* 1024 1024 1024.0))))
	((>= size (* 1024 1024))
	 (format " (%0.2fMB)" (/ size (* 1024 1024.0))))
	((>= size 1024)
	 (format " (%0.2fKB)" (/ size 1024.0))))
       (cond
	((>= speed (* 1024 1024 1024))
	 (format "%0.2f GB/s" (/ speed (* 1024 1024 1024.0))))
	((>= speed (* 1024 1024))
	 (format "%0.2f MB/s" (/ speed (* 1024 1024.0))))
	((>= speed 1024)
	 (format "%0.2f KB/s" (/ speed 1024.0)))
	(t
	 (format "%0.2f B/s" speed)))))))

(emchat-do-in-sxemacs
 (defregexp emchat-curl-filename-regexp (concat emchat-emphasis-url-regexp
					     "/\\([^/].*$\\)")
   "Regexp matching the file part of a URL."))
(emchat-do-in-xemacs
 (defconst emchat-curl-filename-regexp (concat emchat-emphasis-url-regexp
					     "/\\([^/].*$\\)")
   "Regexp matching the file part of a URL."))

(defun emchat-curl-url (url file directory)
  "Download URL to FILE in DIRECTORY."
  (unless (fboundp #'ffi-defun)
    (error 'emchat-curl))
  (add-one-shot-hook 'curl:download&-post-hook emchat-curl-hook-function)
  (curl:download& url (expand-file-name file directory)))

(defun emchat-curl-url-at-point ()
  "Downloads url at point."
  (interactive)
  (when (extentp (extent-at (point)))
    (let* ((dir (if current-prefix-arg
		    (expand-file-name
		     (read-directory-name "Download directory: "))
		  emchat-curl-download-directory))
	   (url (extent-string (extent-at (point))))
	   (fregexp emchat-curl-filename-regexp)
	   file)
      (unless (file-directory-p dir)
	(make-directory-path dir))
      (setq file
	    (and (string-match fregexp url)
		 (not (string=
		       "/"
		       (substring
			(substring url (match-beginning 2) (match-end 2)) -1)))
		 (substring url (match-beginning 2) (match-end 2))))
      (if file
	  (progn
	    (emchat-curl-url url file dir)
	    (setq emchat-curl-file file
		  emchat-curl-dir dir))
	(message "I don't see a file in: %s" url)))))

(defun emchat-curl-url-at-mouse (event)
  "Downloads url at event."
  (interactive "e")
  (when (extentp (extent-at-event event))
    (let* ((dir (if current-prefix-arg
		    (expand-file-name
		     (read-directory-name "Download directory: "))
		  emchat-curl-download-directory))
	   (url (extent-string (extent-at-event event)))
	   (fregexp emchat-curl-filename-regexp)
	   file)
      (unless (file-directory-p dir)
	(make-directory-path dir))
      (setq file
	    (and (string-match fregexp url)
		 (not (string=
		       "/"
		       (substring
			(substring url (match-beginning 2) (match-end 2)) -1)))
		 (substring url (match-beginning 2) (match-end 2))))
      (if file
	  (progn
	    (emchat-curl-url url file dir)
	    (setq emchat-curl-file file
		  emchat-curl-dir dir))
	(message-or-box "I don't see a file in: %s" url)))))

(provide 'emchat-curl)
;;; emchat-curl.el ends here
