;;; emchat-v8.el --- ICQ v8 for emchat.

;; Copyright (C) 2004 - 2011 Steve Youngs, Zajcev Evgeny

;; Author:        Zajcev Evgeny <zevlg@yandex.ru>
;; Maintainer:    Steve Youngs <steve@emchat.org>
;; Created:       Mon Jun  7 16:20:56 MSD 2004
;; Keywords:      emchat ICQ protocol

;; This file is part of EMchat.

;; Redistribution and use in source and binary forms, with or without
;; modification, are permitted provided that the following conditions
;; are met:
;;
;; 1. Redistributions of source code must retain the above copyright
;;    notice, this list of conditions and the following disclaimer.
;;
;; 2. Redistributions in binary form must reproduce the above copyright
;;    notice, this list of conditions and the following disclaimer in the
;;    documentation and/or other materials provided with the distribution.
;;
;; 3. Neither the name of the author nor the names of any contributors
;;    may be used to endorse or promote products derived from this
;;    software without specific prior written permission.
;;
;; THIS SOFTWARE IS PROVIDED BY THE AUTHOR "AS IS" AND ANY EXPRESS OR
;; IMPLIED WARRANTIES, INCLUDING, BUT NOT LIMITED TO, THE IMPLIED
;; WARRANTIES OF MERCHANTABILITY AND FITNESS FOR A PARTICULAR PURPOSE ARE
;; DISCLAIMED.  IN NO EVENT SHALL THE REGENTS OR CONTRIBUTORS BE LIABLE
;; FOR ANY DIRECT, INDIRECT, INCIDENTAL, SPECIAL, EXEMPLARY, OR
;; CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT LIMITED TO, PROCUREMENT OF
;; SUBSTITUTE GOODS OR SERVICES; LOSS OF USE, DATA, OR PROFITS; OR
;; BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY THEORY OF LIABILITY,
;; WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT (INCLUDING NEGLIGENCE
;; OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE OF THIS SOFTWARE, EVEN
;; IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.

;;; Commentary:

;; Testing UIN: 286012066, PASS: 123
;;         UIN: 321529598, PASS: 123
;;
;;  The various network packets that make up the ICQv8 protocol.
;;
;;  See: <http://www.emchat.org/~steve/icq/docs/>
;;       <http://iserverd1.khstu.ru/oscar/index.html>

;;; Code:

(eval-when-compile
  (require 'cl)
  (defvar emchat-add-user-success))

;;;; Constants

;; FLAP channel ids
(defconst emchat-v8-FLAP-HELLO 1)
(defconst emchat-v8-FLAP-SNAC 2)
(defconst emchat-v8-FLAP-ERRORS 3)
(defconst emchat-v8-FLAP-LOGOFF 4)
(defconst emchat-v8-FLAP-PING 5)

(defconst emchat-v8-client-id-string
  "ICQ Inc. - Product of ICQ (TM).2002a.5.37.1.3728.85"
  "A string to identify ourselves as a nice well behaved ICQ client.")

(defconst emchat-v8-FLAP-VER-MAJOR 5)
(defconst emchat-v8-FLAP-VER-MINOR 37)
(defconst emchat-v8-FLAP-VER-LESSER 1)
(defconst emchat-v8-FLAP-VER-BUILD 3828)
(defconst emchat-v8-FLAP-VER-SUBBUILD 85)

(defconst emchat-v8-message-types
  '(normal
    chat-request
    url
    request-auth
    authorize
    added
    pager
    email-express
    email-check
    card
    contact-list)
  "List of valid message types.")

(defconst emchat-v8-valid-incoming-handlers
  '(connected				; args: no
    disconnected			; args: no
    instant-message			; args: uin, msg &optional msg-type, time-stamp
    offline-message			; args: uin, msg &optional msg-type, time-stamp
    end-offline-messages		;
    about-general                       ; general info about user
    about-more
    about-about
    srv-error
    ;; TODO: add more
    )
  "List of valid incoming message handlers.

To bind on to handle incoming part of icq v8 proto.")

(defconst emchat-v8-valid-outgoing-handlers
  '(
    ;; TODO: add some handlers
    )
  "List of valid outgoing message handlers.

To bind on to handle outgoing part of icq v8 proto.")

(defstruct emchat-v8-ctx
  host port				; host/port we are currently connected
  proc					; emchat network connection
  (readingp 0)				; non-0 if we are currently reading from emchat context
  sequence				; current flap sequence

  incoming-buffer				; where incoming data stored
  incoming-handlers
  outgoing-buffer			; buffer for outgoing messages
  outgoing-handles

  state					; current state, connecting, connected
  userinfo				; plist of userinfo '(uin <num> password <passwd>)
  plist					; user defined plist
  )

(defun emchat-v8-ctx-get-prop (ectx prop)
  "Get ECTX's property PROP value."
  (plist-get (emchat-v8-ctx-plist ectx) prop))

(defun emchat-v8-ctx-put-prop (ectx prop val)
  "Put ECTX's property PROP value VAL."
  (setf (emchat-v8-ctx-plist ectx)
	(plist-put (emchat-v8-ctx-plist ectx) prop val)))
(put 'emchat-v8-ctx-put-prop 'lisp-indent-function 2)

(defun emchat-v8-ctx-rm-prop (ectx prop)
  "Remove ECTX's property PROP."
  (setf (emchat-v8-ctx-plist ectx)
	(plist-remprop (emchat-v8-ctx-plist ectx) prop)))

(defun emchat-v8-ctx-get-incoming-handlers (ectx h)
  (plist-get (emchat-v8-ctx-incoming-handlers ectx) h))

(defun emchat-v8-ctx-add-incoming-handler (ectx h f)
  (setf (emchat-v8-ctx-incoming-handlers ectx)
	(plist-put (emchat-v8-ctx-incoming-handlers ectx) h
		   (cons f (emchat-v8-ctx-get-incoming-handlers ectx h)))))

(defun emchat-v8-ctx-del-incoming-handler (ectx h f)
  (setf (emchat-v8-ctx-incoming-handlers ectx)
	(plist-put (emchat-v8-ctx-incoming-handlers ectx) h
		   (delete f (emchat-v8-ctx-get-incoming-handlers ectx h)))))

(defun emchat-v8-ctx-run-incoming-handler (ectx h &rest args)
  (let ((ohs (emchat-v8-ctx-get-incoming-handlers ectx h))
	(ih-arguments args))
    (when ohs
      (declare (special ih-arguments))
      (apply 'run-hook-with-args (cons 'ohs (cons ectx args))))))
(put 'emchat-v8-ctx-run-incoming-handler 'lisp-indent-function 2)

(defun emchat-get-arg (kw)
  (declare (special ih-arguments))
  (plist-get ih-arguments kw))


(defmacro emchat-v8-ctx-msg-to-send (ectx)
  `(emchat-v8-ctx-get-prop ,ectx 'message-to-send))
(defsetf emchat-v8-ctx-msg-to-send (ectx) (msg)
  `(emchat-v8-ctx-put-prop ,ectx 'message-to-send ,msg))

(defmacro emchat-v8-ctx-pause-p (ectx)
  `(emchat-v8-ctx-get-prop ,ectx 'pause-p))
(defsetf emchat-v8-ctx-pause-p (ectx) (pp)
  `(if ,pp
       (progn
	 (emchat-v8-ctx-put-prop ,ectx 'pause-p ,pp)
	 (setf (emchat-v8-ctx-msg-to-send ,ectx) ""))
     (emchat-v8-ctx-rm-prop ,ectx 'pause-p)
     (emchat-v8-send ,ectx (emchat-v8-ctx-msg-to-send ,ectx))
     (setf (emchat-v8-ctx-msg-to-send ,ectx) "")))


(defvar emchat-v8-default-ctx nil
  "Internal variable, do not modify.")

(defvar emchat-v8-connections nil
  "List if emchat connections.
Internal variable, do not modify.")

(defcustom emchat-v8-debug t
  "*Non-nil mean emchat protocol debugging is enabled."
  :type 'boolean)


(defvar emchat-debug-buffer "*emchat-debug*")

(defun emchat-v8-debug (msg &rest args)
  "Print debug MSG."
  (when emchat-v8-debug
    (with-current-buffer
	(get-buffer-create emchat-debug-buffer)
      (save-excursion
	(goto-char (point-max))
	(insert (apply 'format msg args) "\n")))))

(defun emchat-v8-ctx ()
  "Return default EMchat context."
  emchat-v8-default-ctx)

;; FLAP accessors
(defsubst emchat-v8-flap-cid (flap)
  (nth 0 flap))
(defsubst emchat-v8-flap-seq (flap)
  (nth 1 flap))
(defsubst emchat-v8-flap-data (flap)
  (nth 2 flap))

;;; SNAC accessors
(defsubst emchat-v8-snac-family (snac)
  (nth 0 snac))
(defsubst emchat-v8-snac-subtype (snac)
  (nth 1 snac))
(defsubst emchat-v8-snac-flags (snac)
  (nth 2 snac))
(defsubst emchat-v8-snac-rid (snac)
  (nth 3 snac))

;;; TLV accessors
(defsubst emchat-v8-tlv-type (tlv)
  (nth 0 tlv))
(defsubst emchat-v8-tlv-len (tlv)
  (nth 1 tlv))
(defsubst emchat-v8-tlv-val (tlv)
  (nth 2 tlv))
(defsubst emchat-v8-tlv-str (tlv)
  (aref (emchat-v8-tlv-val tlv) 0))
(defsubst emchat-v8-tlv-num (tlv)
  (truncate (emchat-v8-string->number
   (aref (emchat-v8-tlv-val tlv) 0) (truncate (emchat-v8-tlv-len tlv)))))

(defun emchat-v8-util-encrypt (password)
  (let ((tb "\xf3\x26\x81\xc4\x39\x86\xdb\x92\x71\xa3\xb9\xe6\x53\x7a\x95\x7c")
	(res (make-string (length password) ?\x00))
	(idx 0))
    (while (< idx (length password))
      (aset res idx (logxor (aref password idx) (aref tb (mod idx 16))))
      (setq idx (1+ idx)))
    res))

(defun emchat-v8-create-ctx (uin &optional password &rest plist)
  "Create new emchat context using UIN/PASSWORD.
PLIST is ectx properties list"
  (let ((ectx (make-emchat-v8-ctx :state 'disconnected :plist plist)))
    (setf (emchat-v8-ctx-userinfo ectx)
	  (list 'uin uin 'password password))
    ectx))

(defun emchat-v8-connect (ectx &optional server port)
  "Connect to SERVER and PORT."
  (unless server
    (setq server (or (emchat-v8-ctx-host ectx) "login.icq.com")))
  (unless port
    (setq port (or (emchat-v8-ctx-host ectx) 5190)))

  (setf (emchat-v8-ctx-host ectx) server)
  (setf (emchat-v8-ctx-port ectx) port)

  (let* ((proc (open-network-stream (format "emchat-v8-%s:%d" server port) nil
				    server port))
	 (state (emchat-v8-ctx-state ectx)))

    (setf (emchat-v8-ctx-proc ectx) proc)
    (setf (emchat-v8-ctx-sequence ectx) (random 65536))

    ;; Store connection in emchat connections list
    (push ectx emchat-v8-connections)

    (set-process-filter proc 'emchat-v8-proc-filter-proccessing)
    (set-process-sentinel proc 'emchat-v8-proc-sentinel)

    (emchat-v8-debug "Connecting from state=%S" (emchat-v8-ctx-state ectx))

    ;; Adjust ECTX state
    (cond ((or (eq state 'connecting)
	       (eq state 'closed))
	   (setf (emchat-v8-ctx-state ectx) 'connected))
	  (t (setf (emchat-v8-ctx-state ectx) 'connecting)))
    ectx))

(defun emchat-v8-close (ectx)
  "Close ECTX connection."
  (setf (emchat-v8-ctx-state ectx) 'closed)
  (emchat-v8-proc-sentinel (emchat-v8-ctx-proc ectx)))

(defun emchat-v8-send (ectx msg)
  "Using emchat context ECTX, send message MSG."
  (if (not (emchat-v8-ctx-pause-p ectx))
      (process-send-string (emchat-v8-ctx-proc ectx) msg)
    ;; Add MSG to the pending data pending
    (setf (emchat-v8-ctx-msg-to-send ectx)
	  (concat (emchat-v8-ctx-msg-to-send ectx) msg))))
(put 'emchat-v8-send 'lisp-indent-function 1)

(defun emchat-v8-send-flush (ectx)
  "Flush ECTX's output buffer."
  (when (> (length (emchat-v8-ctx-outgoing-buffer ectx)) 0)
    (process-send-string
     (emchat-v8-ctx-proc ectx)
     (emchat-v8-ctx-outgoing-buffer ectx))))

(defun emchat-v8-proc-find (proc)
  "Find emchat context by PROC."
  (let ((ecs emchat-v8-connections))
    (while (and ecs (not (eq (emchat-v8-ctx-proc (car ecs)) proc)))
      (setq ecs (cdr ecs)))
    (car ecs)))

(defun emchat-v8-proc-sentinel (proc &optional event)
  "Sentinel for emchat PROC."
  (let ((ectx (emchat-v8-proc-find proc)))
    (when (emchat-v8-ctx-p ectx)

      (emchat-v8-debug "Removing process %S, state=%S, c-t=%S"
		     proc (emchat-v8-ctx-state ectx) (emchat-v8-ctx-get-prop ectx 'connect-tries))
      (delete-process proc)
      (setf (emchat-v8-ctx-incoming-buffer ectx) "") ;flush input buffer

      (setq emchat-v8-connections (delq ectx emchat-v8-connections))
      (when (eq (emchat-v8-ctx-state ectx) 'connected)
	;; Unexpected disconnection
	(emchat-v8-ctx-run-incoming-handler ectx 'disconnect)))))

(defun emchat-v8-proc-filter-accumulator (proc out)
  "Filter for emchat connection to accumulate data."
  (let ((ectx (emchat-v8-proc-find proc)))
    (when (emchat-v8-ctx-p ectx)
      (setf (emchat-v8-ctx-incoming-buffer ectx)
	    (concat (emchat-v8-ctx-incoming-buffer ectx) out)))))

(defun emchat-v8-proc-filter-proccessing (proc out)
  "Filter for emchat connection."
  (let ((ectx (emchat-v8-proc-find proc)))
    (when (emchat-v8-ctx-p ectx)
      (setf (emchat-v8-ctx-incoming-buffer ectx)
	    (concat (emchat-v8-ctx-incoming-buffer ectx) out))
      (enqueue-eval-event 'emchat-v8-fetch-handle-flaps ectx))))

(defvar emchat-v8-snacv-list
  '((#x01 . #x04)
    (#x13 . #x04)
    (#x02 . #x01)
    (#x03 . #x01)
    (#x15 . #x01)
    (#x04 . #x01)
    (#x06 . #x01)
    (#x09 . #x01)
    (#x0A . #x01)
    (#x0B . #x01)))

(defconst emchat-v8-snac-list
  '(
    ;; Family 1, Service
    ((#x01 . #x01) "SRV_GEN_ERROR"      emchat-v8-snac-srv-err)
    ((#x01 . #x02) "CLI_READY"          nil)
    ((#x01 . #x03) "SRV_FAMILIES"       emchat-v8-snac-srv-families)
    ((#x01 . #x06) "CLI_RATESREQUEST"   emchat-v8-snac-cli-ratesrequest)
    ((#x01 . #x07) "SRV_RATES"          emchat-v8-snac-srv-rates)
    ;((#x01 . #x08) "CLI_ACKRATES"       nil) ; handled in SRV_RATES
    ((#x01 . #x0A) "SRV_RATE_CHANGE"    emchat-v8-snac-srv-rate-change)
    ((#x01 . #x0B) "SRV_SERVERPAUSE"    emchat-v8-snac-srv-pause)
    ((#x01 . #x0C) "CLI_ACKSERVERPAUSE" nil)
    ((#x01 . #x0E) "CLI_REQINFO"        nil)
    ((#x01 . #x0F) "SRV_ONLINExINFO"    emchat-v8-snac-srv-onlinexinfo)
    ((#x01 . #x12) "SRV_MIGRATIONREQ"   emchat-v8-snac-srv-migrate)
    ((#x01 . #x13) "SRV_MOTD"           emchat-v8-snac-srv-motd)
    ((#x01 . #x15) "SRV_REPLYINFO"      emchat-v8-snac-srv-replyinfo)
    ((#x01 . #x17) "CLI_FAMILIES"       nil)
    ((#x01 . #x18) "SRV_FAMILIES2"      emchat-v8-snac-srv-families2)
    ((#x01 . #x1E) "CLI_SETSTATUS"      nil)
    ((#x01 . #x21) "BART_REQUEST"       emchat-v8-snac-bart-request)
    ;; Family 2, Location
    ((#x02 . #x01) "SRV_GEN_ERROR"      emchat-v8-snac-srv-err)
    ((#x02 . #x02) "CLI_REQLOCATION"    nil)
    ((#x02 . #x03) "SRV_REPLYLOCATION"  emchat-v8-snac-srv-reply-location)
    ((#x02 . #x04) "CLI_SETUSERINFO"    nil)
    ;; Family 3, Buddy lists (looks like AOL has turned this off)
    ((#x03 . #x01) "SRV_CONTACTERR"     emchat-v8-snac-srv-contact-err)
    ((#x03 . #x02) "CLI_REQBUDDY"       nil)
    ((#x03 . #x03) "SRV_REPLYBUDDY"     emchat-v8-snac-srv-reply-buddy)
    ((#x03 . #x04) "CLI_ADDCONTACT"     nil)
    ((#x03 . #x05) "CLI_REMCONTACT"     nil)
    ((#x03 . #x0A) "SRV_REFUSE"         emchat-v8-snac-srv-cont-refused)
    ((#x03 . #x0B) "SRV_USERONLINE"     emchat-v8-snac-srv-user-online)
    ((#x03 . #x0C) "SRV_USEROFFLINE"    emchat-v8-snac-srv-user-offline)
    ;; Family 4, ICBM
    ((#x04 . #x01) "SRV_ICBMERR"        emchat-v8-snac-srv-icbm-err)
    ((#x04 . #x02) "CLI_SETICBM"        nil)
    ((#x04 . #x04) "CLI_REQICBM"        nil)
    ((#x04 . #x05) "SRV_REPLYICBM"      emchat-v8-snac-srv-reply-icbm)
    ((#x04 . #x06) "CLI_SENDMSG"        nil)
    ((#x04 . #x07) "SRV_RECVMSG"        emchat-v8-snac-srv-recv-msg)
    ((#x04 . #x0A) "SRV_MISSED_ICBM"    emchat-v8-snac-srv-missed-icbm)
    ((#x04 . #x0B) "SRV/CLI_ACKMSG"     emchat-v8-snac-srv-ack-msg)
    ((#x04 . #x0C) "SRV_SRVACKMSG"      emchat-v8-snac-srv-srv-ack-msg)
    ;; Family 5, Advertising NOT USED (thank god!)
    ;; Family 6, Invitation (Invite somebody to AIM) erm, EMchat won't!!
    ;; Family 7, Admin (AIM)
    ;; Family 8, Popup Notices (AIM)
    ;; Family 9, Privacy
    ((#x09 . #x01) "SRV_GEN_ERROR"      emchat-v8-snac-srv-err)
    ((#x09 . #x02) "CLI_REQBOS"         nil)
    ((#x09 . #x03) "SRV_REPLYBOS"       emchat-v8-snac-srv-reply-bos)
    ((#x09 . #x05) "CLI_ADDVISIBLE"     emchat-v8-snac-cli-addvisible)
    ((#x09 . #x06) "CLI_REMVISIBLE"     emchat-v8-snac-cli-remvisible)
    ((#x09 . #x07) "CLI_ADDINVISIBLE"   emchat-v8-snac-cli-addinvisible)
    ((#x09 . #x08) "CLI_REMINVISIBLE"   emchat-v8-snac-cli-reminvisible)
    ;; Family a [10], User Lookup (AIM and no longer used)
    ;; Family b [11], Usage Stats Interval
    ((#x0B . #x01) "SRV_GEN_ERROR"      emchat-v8-snac-srv-err)
    ((#x0B . #x02) "SRV_SETINTERVAL"    emchat-v8-snac-srv-set-interval)
    ;; Family c [12], Translation Service (deprecated (AIM))
    ;; Family d [13], Chat Navigation (AIM)
    ;; Family e [14], Chat Service (AIM)
    ;; Family f [15], Directory Search (AIM)
    ;; Family 10 [16], Server Stored Buddy Icons
    ;; Family 13 [19], Server Side Info
    ((#x13 . #x01) "SRV_GEN_ERROR" emchat-v8-snac-srv-err)
    ((#x13 . #x02) "CLI_SSI_RIGHTS_REQUEST" emchat-v8-snac-cli-ssi-right-request)
    ((#x13 . #x03) "SRV_SSI_RIGHTS_REPLY" emchat-v8-snac-srv-ssi-rights-reply)
    ((#x13 . #x04) "CLI_SSI_REQUEST" emchat-v8-snac-cli-ssi-request)
    ((#x13 . #x05) "CLI_SSI_CHECKOUT" emchat-v8-snac-cli-ssi-checkout)
    ((#x13 . #x06) "SRV_SSI_REPLY" emchat-v8-snac-srv-ssi-reply)
    ((#x13 . #x07) "CLI_SSI_ACTIVATE" emchat-v8-snac-cli-ssi-activate)
    ((#x13 . #x08) "CLI_SSI_ADD" emchat-v8-snac-cli-ssi-add) ; broken
    ((#x13 . #x09) "CLI_SSI_UPDATE" emchat-v8-snac-cli-ssi-update) ; not implimented yet
    ((#x13 . #x0A) "CLI_SSI_DELETE" emchat-v8-snac-cli-ssi-delete) ; not implimented yet
    ((#x13 . #x0E) "SRV_SSI_MOD_ACK" emchat-v8-snac-srv-ssi-mod-ack)
    ((#x13 . #x0F) "SRV_SSI_UP_TO_DATE" emchat-v8-snac-srv-ssi-up-to-date)
    ((#x13 . #x11) "CLI_SSI_EDIT_BEGIN" emchat-v8-snac-cli-ssi-edit-begin)
    ((#x13 . #x12) "CLI_SSI_EDIT_END" emchat-v8-snac-cli-ssi-edit-end)
    ((#x13 . #x14) "CLI_SSI_FUTURE_AUTH_GRANT" emchat-v8-snac-cli-ssi-future-auth-grant)
    ((#x13 . #x15) "SRV_SSI_FUTURE_AUTH_GRANTED" emchat-v8-snac-srv-ssi-future-auth-granted)
    ((#x13 . #x16) "CLI_SSI_DEL_YOURSELF" emchat-v8-snac-cli-ssi-del-yourself)
    ((#x13 . #x18) "CLI_SSI_SEND_AUTH_REQUEST" emchat-v8-snac-cli-ssi-send-auth-request)
    ((#x13 . #x19) "SRV_SSI_AUTH_REQUEST" emchat-v8-snac-srv-ssi-auth-request)
    ((#x13 . #x1A) "CLI_SSI_AUTH_REPLY" emchat-v8-snac-cli-ssi-auth-reply)
    ((#x13 . #x1B) "SRV_SSI_AUTH_REPLY" emchat-v8-snac-srv-ssi-auth-reply)
    ((#x13 . #x1C) "SRV_SSI_YOU_WERE_ADDED" emchat-v8-snac-srv-ssi-you-were-added)
    ;; Family 15 [21], ICQ Specific
    ((#x15 . #x01) "SRV_TOICQERR"       emchat-v8-snac-srv-toicqerr)
    ;; This is like a family within a family... all the SNACS are #x15,#x02
    ;; but they also have types and subtypes to differentiate between them.
    ;; So we really don't need a function here.
    ((#x15 . #x02) "CLI_TOICQSRV"       nil)
    ((#x15 . #x03) "SRV_FROMICQSRV"     emchat-v8-snac-srv-fromicqsrv)
    ;; Family 17 [23], Authorisation/Registration
    ((#x17 . #x01) "SRV_REGREFUSED"     emchat-v8-snac-srv-regrefused)
    ((#x17 . #x05) "SRV_NEWUIN"         emchat-v8-snac-srv-newuin)
    ((#x17 . #x04) "CLI_REGISTERUSER"   nil)

    ((0 . 0)    "unknown"            nil))
  "List of SNAC specs.")

(defun emchat-v8-recv-message (ectx h-type data)
  "Receive message."
  (let* ((uin (emchat-v8-fetch-uin data))
	 (msg (emchat-v8-parse-message data
		(list [2 number-swap]	;year
		      [1 number]	;month
		      [1 number]	;day
		      [1 number]	;hour
		      [1 number]	;minutes
		      )))
	 (tmsg (emchat-v8-parse-message data
		 (list [2 number-swap]
		       [2 number])))
	 (ctext (substring (aref data 0) 0 (1- (length (aref data 0)))))
	 (time-stamp (list (truncate (nth 0 msg))
			   (truncate (nth 1 msg))
			   (truncate (nth 2 msg))
			   (truncate (nth 3 msg))
			   (truncate (nth 4 msg)))))

    (emchat-v8-debug "Recv-message: '%S', tmsg='%S'" ctext tmsg)

    ;; XXX 'normal
    (emchat-v8-ctx-run-incoming-handler ectx h-type
      :uin uin :msg ctext :msg-type 'normal :time-stamp time-stamp)))

(defmacro emchat-v8-get-alist-value (code alist)
  `(cdr (assq ,code ,alist)))

(defconst emchat-v8-countries-alist
  '((0 . "Undefined")
    (1 . "USA")
    (101 . "Anguilla")
    (102 . "Antigua")
    (1021 . "Antigua & Barbuda")
    (103 . "Bahamas")
    (104 . "Barbados")
    (105 . "Bermuda")
    (106 . "British Virgin Islands")
    (107 . "Canada")
    (108 . "Cayman Islands")
    (109 . "Dominica")
    (110 . "Dominican Republic")
    (111 . "Grenada")
    (112 . "Jamaica")
    (113 . "Montserrat")
    (114 . "Nevis")
    (1141 . "Saint Kitts and Nevis")
    (115 . "St. Kitts")
    (116 . "St. Vincent & the Grenadines")
    (117 . "Trinidad & Tobago")
    (118 . "Turks & Caicos Islands")
    (120 . "Barbuda")
    (121 . "Puerto Rico")
    (122 . "Saint Lucia")
    (123 . "Virgin Islands (USA)")
    (178 . "Canary Islands")
    (20 . "Egypt")
    (212 . "Morocco")
    (213 . "Algeria")
    (216 . "Tunisia")
    (218 . "Libyan Arab Jamahiriya")
    (220 . "Gambia")
    (221 . "Senegal")
    (222 . "Mauritania")
    (223 . "Mali")
    (224 . "Guinea")
    (225 . "Cote d'Ivoire")
    (226 . "Burkina Faso")
    (227 . "Niger")
    (228 . "Togo")
    (229 . "Benin")
    (230 . "Mauritius")
    (231 . "Liberia")
    (232 . "Sierra Leone")
    (233 . "Ghana")
    (234 . "Nigeria")
    (235 . "Chad")
    (236 . "Central African Republic")
    (237 . "Cameroon")
    (238 . "Cape Verde Islands")
    (239 . "Sao Tome & Principe")
    (240 . "Equatorial Guinea")
    (241 . "Gabon")
    (242 . "Congo, (Rep. of the)")
    (243 . "Congo, Democratic Republic of")
    (244 . "Angola")
    (245 . "Guinea-Bissau")
    (246 . "Diego Garcia")
    (247 . "Ascension Island")
    (248 . "Seychelles")
    (249 . "Sudan")
    (250 . "Rwanda")
    (251 . "Ethiopia")
    (252 . "Somalia")
    (253 . "Djibouti")
    (254 . "Kenya")
    (255 . "Tanzania")
    (256 . "Uganda")
    (257 . "Burundi")
    (258 . "Mozambique")
    (260 . "Zambia")
    (261 . "Madagascar")
    (262 . "Reunion Island")
    (263 . "Zimbabwe")
    (264 . "Namibia")
    (265 . "Malawi")
    (266 . "Lesotho")
    (267 . "Botswana")
    (268 . "Swaziland")
    (269 . "Mayotte Island")
    (2691 . "Comoros")
    (27 . "South Africa")
    (290 . "St. Helena")
    (291 . "Eritrea")
    (297 . "Aruba")
    (298 . "Faeroe Islands")
    (299 . "Greenland")
    (30 . "Greece")
    (31 . "Netherlands")
    (32 . "Belgium")
    (33 . "France")
    (34 . "Spain")
    (350 . "Gibraltar")
    (351 . "Portugal")
    (352 . "Luxembourg")
    (353 . "Ireland")
    (354 . "Iceland")
    (355 . "Albania")
    (356 . "Malta")
    (357 . "Cyprus")
    (358 . "Finland")
    (359 . "Bulgaria")
    (36 . "Hungary")
    (370 . "Lithuania")
    (371 . "Latvia")
    (372 . "Estonia")
    (373 . "Moldova, Republic of")
    (374 . "Armenia")
    (375 . "Belarus")
    (376 . "Andorra")
    (377 . "Monaco")
    (378 . "San Marino")
    (379 . "Vatican City")
    (380 . "Ukraine")
    (381 . "Yugoslavia")
    (3811 . "Yugoslavia - Serbia")
    (382 . "Yugoslavia - Montenegro")
    (385 . "Croatia")
    (386 . "Slovenia")
    (387 . "Bosnia & Herzegovina")
    (389 . "Macedonia (F.Y.R.O.M.)")
    (39 . "Italy")
    (40 . "Romania")
    (41 . "Switzerland")
    (4101 . "Liechtenstein")
    (42 . "Czech Republic")
    (4201 . "Slovakia")
    (43 . "Austria")
    (44 . "United Kingdom")
    (441 . "Wales")
    (442 . "Scotland")
    (45 . "Denmark")
    (46 . "Sweden")
    (47 . "Norway")
    (48 . "Poland")
    (49 . "Germany")
    (500 . "Falkland Islands")
    (501 . "Belize")
    (502 . "Guatemala")
    (503 . "El Salvador")
    (504 . "Honduras")
    (505 . "Nicaragua")
    (506 . "Costa Rica")
    (507 . "Panama")
    (508 . "St. Pierre & Miquelon")
    (509 . "Haiti")
    (51 . "Peru")
    (52 . "Mexico")
    (53 . "Cuba")
    (54 . "Argentina")
    (55 . "Brazil")
    (56 . "Chile, Republic of")
    (57 . "Colombia")
    (58 . "Venezuela")
    (590 . "Guadeloupe")
    (5901 . "French Antilles")
    (5902 . "Antilles")
    (591 . "Bolivia")
    (592 . "Guyana")
    (593 . "Ecuador")
    (594 . "French Guyana")
    (595 . "Paraguay")
    (596 . "Martinique")
    (597 . "Suriname")
    (598 . "Uruguay")
    (599 . "Netherlands Antilles")
    (60 . "Malaysia")
    (61 . "Australia")
    (6101 . "Cocos-Keeling Islands")
    (6102 . "Cocos (Keeling) Islands")
    (62 . "Indonesia")
    (63 . "Philippines")
    (64 . "New Zealand")
    (65 . "Singapore")
    (66 . "Thailand")
    (670 . "Saipan Island")
    (6701 . "Rota Island")
    (6702 . "Tinian Island")
    (671 . "Guam, US Territory of")
    (672 . "Christmas Island")
    (6722 . "Norfolk Island")
    (673 . "Brunei")
    (674 . "Nauru")
    (675 . "Papua New Guinea")
    (676 . "Tonga")
    (677 . "Solomon Islands")
    (678 . "Vanuatu")
    (679 . "Fiji")
    (680 . "Palau")
    (681 . "Wallis & Futuna Islands")
    (682 . "Cook Islands")
    (683 . "Niue")
    (684 . "American Samoa")
    (685 . "Western Samoa")
    (686 . "Kiribati")
    (687 . "New Caledonia")
    (688 . "Tuvalu")
    (689 . "French Polynesia")
    (690 . "Tokelau")
    (691 . "Micronesia, Federated States of")
    (692 . "Marshall Islands")
    (7 . "Russia")
    (705 . "Kazakhstan")
    (706 . "Kyrgyzstan")
    (708 . "Tajikistan")
    (709 . "Turkmenistan")
    (711 . "Uzbekistan")
    (81 . "Japan")
    (82 . "Korea, South")
    (84 . "Viet Nam")
    (850 . "Korea, North")
    (852 . "Hong Kong")
    (853 . "Macau")
    (855 . "Cambodia")
    (856 . "Laos")
    (86 . "China")
    (880 . "Bangladesh")
    (886 . "Taiwan")
    (90 . "Turkey")
    (91 . "India")
    (92 . "Pakistan")
    (93 . "Afghanistan")
    (94 . "Sri Lanka")
    (95 . "Myanmar")
    (960 . "Maldives")
    (961 . "Lebanon")
    (962 . "Jordan")
    (963 . "Syrian Arab Republic")
    (964 . "Iraq")
    (965 . "Kuwait")
    (966 . "Saudi Arabia")
    (967 . "Yemen")
    (968 . "Oman")
    (971 . "United Arabian Emirates")
    (972 . "Israel")
    (973 . "Bahrain")
    (974 . "Qatar")
    (975 . "Bhutan")
    (976 . "Mongolia")
    (977 . "Nepal")
    (98 . "Iran (Islamic Republic of)")
    (994 . "Azerbaijan")
    (995 . "Georgia")
    (9999 . "other")))

(defmacro emchat-v8-get-country (code)
  "Return country name according to country CODE."
  `(emchat-v8-get-alist-value ,code emchat-v8-countries-alist))

(defconst emchat-v8-gender-alist
   '((0 . "unspecified")
     (1 . "female")
     (2 . "male")))

(defmacro emchat-v8-get-gender (code)
  "Return gender name according to specified gender CODE."
  `(emchat-v8-get-alist-value ,code emchat-v8-gender-alist))

(defconst emchat-v8-languages-alist
  '((0 . "not specified")
    (1 . "Arabic")
    (2 . "Bhojpuri")
    (3 . "Bulgarian")
    (4 . "Burmese")
    (5 . "Cantonese")
    (6 . "Catalan")
    (7 . "Chinese")
    (8 . "Croatian")
    (9 . "Czech")
    (10 . "Danish")
    (11 . "Dutch")
    (12 . "English")
    (13 . "Esperanto")
    (14 . "Estonian")
    (15 . "Farsi")
    (16 . "Finnish")
    (17 . "French")
    (18 . "Gaelic")
    (19 . "German")
    (20 . "Greek")
    (21 . "Hebrew")
    (22 . "Hindi")
    (23 . "Hungarian")
    (24 . "Icelandic")
    (25 . "Indonesian")
    (26 . "Italian")
    (27 . "Japanese")
    (28 . "Khmer")
    (29 . "Korean")
    (30 . "Lao")
    (31 . "Latvian")
    (32 . "Lithuanian")
    (33 . "Malay")
    (34 . "Norwegian")
    (35 . "Polish")
    (36 . "Portuguese")
    (37 . "Romanian")
    (38 . "Russian")
    (39 . "Serbian")
    (40 . "Slovak")
    (41 . "Slovenian")
    (42 . "Somali")
    (43 . "Spanish")
    (44 . "Swahili")
    (45 . "Swedish")
    (46 . "Tagalog")
    (47 . "Tatar")
    (48 . "Thau")
    (49 . "Turkish")
    (50 . "Ukarinian")
    (51 . "Urdu")
    (52 . "Vietnamese")
    (53 . "Yiddish")
    (54 . "Yoruba")
    (55 . "Afriaans")
    (56 . "Bosnian")
    (57 . "Persian")
    (58 . "Albanian")
    (59 . "Armenian")
    (60 . "Punjabi")
    (61 . "Chamorro")
    (62 . "Mongolian")
    (63 . "Mandarin")
    (64 . "Taiwanese")
    (65 . "Macedonian")
    (66 . "Sindhi")
    (67 . "Welsh")
    (68 . "Azerbaijani")
    (69 . "Kurdish")
    (70 . "Gujarati")
    (71 . "Tamil")
    (72 . "Belorussian")
    (255 . "other")))

(defmacro emchat-v8-get-language (code)
  "Return language name according to language CODE."
  `(emchat-v8-get-alist-value ,code emchat-v8-languages-alist))

(defconst emchat-v8-marital-alist
  '((0 . "not specified")
    (10 . "single")
    (11 . "in a long-term relationship")
    (12 . "engaged")
    (20 . "married")
    (30 . "divorced")
    (31 . "separated")
    (40 . "widowed")))

(defmacro emchat-v8-get-marital (code)
  "Return marital name according to marital CODE."
  `(emchat-v8-get-alist-value ,code emchat-v8-marital-alist))

;;;; SRV

;;; Family #x13 (SSI - Server Side Info)

;;; FIXME: This should most likely be used to set some max values in
;;; EMchat, but I'm not that concerned seeing as though max contacts
;;; is 3000.
(defun emchat-v8-snac-srv-ssi-rights-reply (ectx data &optional flags)
  "SRV_SSI_RIGHTS_REPLY, SNAC(#x13, #x03)."
  (let* ((tlvs (emchat-v8-fetch-tlvs data))
	 (limits (emchat-v8-tlv-val (emchat-v8-tlv-get tlvs 4)))
	 (mconts (emchat-v8-fetch-word limits))  ; max contacts
	 (mgrps (emchat-v8-fetch-word limits))   ; max groups
	 (mvconts (emchat-v8-fetch-word limits)) ; max visible contacts
	 (miconts (emchat-v8-fetch-word limits)) ; max invisible contacts
	 (mvibit (emchat-v8-fetch-word limits))  ; max vis/invis bitmasks
	 (mpres (emchat-v8-fetch-word limits))   ; max presence info fields
	 (mignores                               ; max ignore entries
	  (progn
	    (dotimes (i 8)
	      (emchat-v8-fetch-word limits))
	    (emchat-v8-fetch-word limits)))
	 ;; rest of the thing is ignored for now (unknown stuff)
	 )
    ;; not sure what to do with this yet, lets just log it for now
    (emchat-v8-debug "SRV_SSI_RIGHTS_REPLY: Max number of...
Contacts=%d
Groups=%d
Visible Contacts=%d
Invisible Contacts=%d
Vis/Invis Bitmasks=%d
Presense Info Fields=%d
Ignore List Entries=%d"
		     mconts mgrps mvconts miconts mvibit mpres mignores)))

(defvar emchat-v8-ssi-count nil
  "A place to store SSI entry count when it spans multiple packets.")

;;; FIXME: handle BARTs
(defun emchat-v8-snac-srv-ssi-reply (ectx data &optional flags)
  "SRV_SSI_REPLY, SNAC(#x13, #x06)."
  (let* ((count (progn
		  (emchat-v8-fetch-byte data) ; skip, just a version number
		  (emchat-v8-fetch-word data)))
	 lastupd)
    (push count emchat-v8-ssi-count)
    (dotimes (i count)
      (let* ((name (emchat-v8-fetch-bstr data))
	     (group (emchat-v8-fetch-word data))
	     (id (emchat-v8-fetch-word data))
	     (type (emchat-v8-fetch-word data))
	     (tlvs-len (emchat-v8-fetch-word data))
	     (tlvs (emchat-v8-fetch-n-tlvs data tlvs-len))
	     ttype tlen tval)
	(while tlvs
	  (setq ttype (emchat-v8-tlv-type (car tlvs)))
	  (setq tlen (emchat-v8-tlv-len (car tlvs)))
	  (setq tval (emchat-v8-tlv-val (car tlvs)))
	  (when (and (eq ttype #x131) (not (eq type #x14)))
	    (emchat-world-sync-ssi-maybe name group id tval))
	  (emchat-v8-debug "SSI_REPLY_TLV:
name=%S group=%S id=%S type=%S
ttype=%S tlength=%S tvalue=%S"
			   name group id type ttype tlen tval)
	  (setq tlvs (cdr tlvs)))))
    (with-current-buffer (find-file-noselect emchat-world-rc-filename)
      (when (buffer-modified-p)
	(save-buffer)))
    (when (zerop (logand flags 1))
      (when (> (length emchat-v8-ssi-count) 1)
	(setq count (apply #'+ emchat-v8-ssi-count)))
      (setq emchat-v8-ssi-count nil)
      (setq lastupd (emchat-v8-fetch-time data))
      (emchat-world-update-world-count count lastupd)
      (emchat-v8-debug "SSI_REPLY: count=%S lastupd=%S" count lastupd))))

(defun emchat-v8-snac-srv-ssi-mod-ack (ectx data &optional flags)
  "Server ack for buddy add/del/mod.  SNAC(#x13, #x0E)."
  (let ((retcodes '((#x0000 . "Success")
		    (#x0002 . "Item not found")
		    (#x0003 . "Item already exists")
		    (#x000A . "Invalid data")
		    (#x000C . "Limit exceeded")
		    (#x000D . "Can't add ICQ contact to AIM list")
		    (#x000E . "Authorisation required"))))
    (setq emchat-add-user-success nil)
    (while data
      (let* ((code (emchat-v8-fetch-word data))
	     (retstr (cdr (assq code retcodes))))
	(emchat-v8-debug "SRV_SSI_MOD_ACK: %s" retstr)
	(if (zerop code)
	    (emchat-v8-ctx-run-incoming-handler 'ectx 'new-user)
	  (emchat-log-error "SSI Modifications Error: %s" retstr))))))

(defun emchat-v8-snac-srv-ssi-up-to-date (ectx data &optional flags)
  "Local copy of SSI is up to date.  SNAC(#x13, #x0F).

This doesn't actually do anything because there isn't any point.  It
sends back the mod time and item count, but you only get this packet
if your local SSI has the same mod time and item count."
  (emchat-log-info "Local copy of SSI is up to date!")
  (emchat-v8-debug "SRV_SSI_UP_TO_DATE: local SSI is up to date."))


(defun emchat-v8-snac-srv-ssi-auth-reply (ectx data &optional flags)
  "SRV_SSI_AUTH_REPLY, SNAC (#x13 . #x1B)"
  (let ((buin (emchat-v8-fetch-buin data))
	(accept (emchat-v8-fetch-byte data))
	(reason (emchat-v8-fetch-bstr data)))
    (emchat-v8-debug "AUTHREPLY: from %d %s"
		   buin
		   (if (zerop accept)
		       "rejected"
		     "accepted"))
    (if (zerop accept)
	(emchat-v8-ctx-run-incoming-handler ectx 'auth-reject
	  :uin buin :msg reason :msg-type 'auth-reject)
      (emchat-v8-ctx-run-incoming-handler ectx 'auth-accept
	:uin buin :msg reason :msg-type 'auth-accept))))

(defun emchat-v8-snac-srv-ssi-you-were-added (ectx data &optional flags)
  "SRV_ADDEDYOU, SNAC (#x13 . #x1C)"
  (let ((buin (emchat-v8-fetch-buin data))
	(msg "has added you to their contact list."))
    (emchat-v8-debug "ADDEDYOU: by %d" buin)
    (emchat-v8-ctx-run-incoming-handler ectx 'added-you
	:uin buin :msg msg :msg-type 'added)))

(defun emchat-v8-snac-srv-ssi-future-auth-granted (ectx data &optional flags)
  "Future auth has been given.  SNAC(#x13, #x15)."
  (let ((uin (emchat-v8-fetch-buin data))
	(reason (emchat-v8-fetch-bstr data)))
    (emchat-v8-debug "SRV_SSI_FUTURE_AUTH_GRANTED: %s grants auth because: %s"
		     uin reason)
    (emchat-v8-ctx-run-incoming-handler ectx 'auth-accept
      :uin uin :msg reason :msg-type 'auth-accept)))

(defun emchat-v8-snac-srv-ssi-auth-request (ectx data &optional flags)
  "SRV_SSI_AUTH_REQUEST, SNAC (#x13 . #x19)"
  (let ((buin (emchat-v8-fetch-buin data))
	(reason (emchat-v8-fetch-bstr data)))
    (emchat-v8-debug "AUTHREQ: from %d Reason: %S" buin reason)
    (emchat-v8-ctx-run-incoming-handler ectx 'auth-request
      :uin buin :msg reason :msg-type 'auth-request)))

;;; Family #x15 (Old ICQ)

;;; FIXME: complete this, a fair bit is still missing.  Might need
;;; splitting up into several functions
(defun emchat-v8-snac-srv-fromicqsrv (ectx data &optional flags)
  "SRV_FROMICQSRV, SNAC(#x15, #x03)"
  (let* ((tlv1 (emchat-v8-tlv-val (emchat-v8-fetch-single-tlv data)))
	 (len (emchat-v8-fetch-word-le tlv1))
	 (my-uin (emchat-v8-fetch-uin tlv1))
	 (type (emchat-v8-fetch-word-le tlv1))
	 (seq (emchat-v8-fetch-word-le tlv1))
	 (his-uin (cdr (assq seq (emchat-v8-ctx-get-prop ectx 'about-query-seq-uin)))))
    ;; Shutup compiler
    (setq len len my-uin my-uin)

    (emchat-v8-debug "FROMICQSRV: UIN: %S, Type=%d, seq=%d, tlv1=%S, data=%S"
		   his-uin type seq tlv1 data)

    (cond ((= type 65)
	   (emchat-v8-recv-message ectx 'offline-message tlv1))

	  ((= type 66)
	   ;; Ack end of offline messages
	   (emchat-v8-send ectx
	     (emchat-v8-pack-meta-snac ectx 62))
	   (emchat-v8-ctx-run-incoming-handler ectx 'end-offline-messages))

	  ((= type 2010)
	   (let ((subtype (emchat-v8-fetch-word-le tlv1))
		 (result (emchat-v8-fetch-byte tlv1)))
	     (emchat-v8-debug "SRV_META: Subtype=%d, result=%d, tlv1=%S" subtype result tlv1)

	     (when (= result 10)
	       (cond ((= subtype 200)
		      ;; -- Start --
		      ;; GENERAL
		      (let ((nick (emchat-v8-fetch-lnts tlv1))
			    (fname (emchat-v8-fetch-lnts tlv1))
			    (sname (emchat-v8-fetch-lnts tlv1))
			    (email (emchat-v8-fetch-lnts tlv1))
			    (city (emchat-v8-fetch-lnts tlv1))
			    (state (emchat-v8-fetch-lnts tlv1))
			    (phone (emchat-v8-fetch-lnts tlv1))
			    (fax (emchat-v8-fetch-lnts tlv1))
			    (street (emchat-v8-fetch-lnts tlv1))
			    (cellular (emchat-v8-fetch-lnts tlv1))
			    (zip (emchat-v8-fetch-lnts tlv1))
			    (country (emchat-v8-get-country
				      (emchat-v8-fetch-word-le tlv1)))
			    ;; The timezone the user lives in, as
			    ;; multiples of 30minutes relative to UTC.
			    (tz (emchat-v8-fetch-byte tlv1))
			    (flags (emchat-v8-fetch-byte tlv1))
			    (web-ind (= (emchat-v8-fetch-word-le tlv1) 1)))

			(emchat-v8-debug "GENERAL-Info: UIN: %d, Nick: %s"
				       his-uin nick)

			(emchat-v8-ctx-run-incoming-handler ectx 'about-general
			  :uin his-uin
			  :nick nick :first-name fname :second-name sname
			  :email email :city city :state state :phone phone
			  :fax fax :street street :cellular cellular
			  :zip-code zip :country country :tz tz
			  :flags flags :web-indicator web-ind)))

		     ((= subtype 210)
		      ;; WORK
		      ;; Hey!  Does anybody actually care about this? --lg
		      )
		     ((= subtype 220)
		      ;; MORE
		      (let ((age (emchat-v8-fetch-word-le tlv1))
			    (gender (emchat-v8-get-gender
				     (emchat-v8-fetch-byte tlv1)))
			    (homepage (emchat-v8-fetch-lnts tlv1))
			    (birth-year (emchat-v8-fetch-word-le tlv1))
			    (birth-month (emchat-v8-fetch-byte tlv1))
			    (birth-day (emchat-v8-fetch-byte tlv1))
			    (lang1 (emchat-v8-get-language
				    (emchat-v8-fetch-byte tlv1)))
			    (lang2 (emchat-v8-get-language
				    (emchat-v8-fetch-byte tlv1)))
			    (lang3 (emchat-v8-get-language
				    (emchat-v8-fetch-byte tlv1)))
			    (ocity (progn
				     (emchat-v8-fetch-word-le tlv1) ; skip
				     (emchat-v8-fetch-lnts tlv1)))
			    (ostate (emchat-v8-fetch-lnts tlv1))
			    (ocountry (emchat-v8-get-country
				       (emchat-v8-fetch-word-le tlv1)))
			    (marital (emchat-v8-get-marital
				      (emchat-v8-fetch-word-le tlv1))))
			(emchat-v8-ctx-run-incoming-handler ectx 'about-more
			  :uin his-uin :age age :gender gender :homepage homepage
			  :birth-year birth-year :birth-month birth-month
			  :birth-day birth-day :lang1 lang1 :lang2 lang2
			  :lang3 lang3 :ocity ocity :ostate ostate
			  :ocountry ocountry :marital marital)))

		     ((= subtype 235)
		      ;; EMAILS
		      )
		     ((= subtype 270)
		      ;; META270
		      )
		     ((= subtype 230)
		      ;; ABOUT
		      (let ((about-user (emchat-v8-fetch-lnts tlv1)))
			(emchat-v8-debug "USER ABOUT-Info: %S" about-user)
			(emchat-v8-ctx-run-incoming-handler ectx 'about-about
			  :uin his-uin :about about-user)))

		     ((= subtype 240)
		      ;; INTEREST
		      )
		     ((= subtype 250)
		      ;; BACKGROUND
		      )

		     ((= subtype 420)
		      ;; Search found a user
		      (let ((uin (progn (emchat-v8-fetch-word-le tlv1)
					(emchat-v8-fetch-uin tlv1)))
			    (nick (emchat-v8-fetch-lnts tlv1))
			    (first-name (emchat-v8-fetch-lnts tlv1))
			    (last-name (emchat-v8-fetch-lnts tlv1))
			    (email (emchat-v8-fetch-lnts tlv1))
			    (auth (emchat-v8-fetch-byte tlv1))
			    (status (emchat-v8-fetch-word-le tlv1))
			    (gender (emchat-v8-get-gender
				     (emchat-v8-fetch-byte tlv1)))
			    (age (emchat-v8-fetch-word-le tlv1))
			    (missed (emchat-v8-fetch-word-le tlv1)))
			(emchat-v8-debug "FOUND USER: UIN=%d, Nick=%s, missed=%d" uin nick missed)
			(emchat-v8-ctx-run-incoming-handler ectx 'search-found
			  :uin uin :nick nick
			  :first-name first-name :last-name last-name
			  :email email :auth auth :status status :gender gender
			  :age age :missed missed)))

		     ((= subtype 430)
		      ;; Search found a user (the last match from the search)
		      (let ((uin (progn (emchat-v8-fetch-word-le tlv1)
					(emchat-v8-fetch-uin tlv1)))
			    (nick (emchat-v8-fetch-lnts tlv1))
			    (first-name (emchat-v8-fetch-lnts tlv1))
			    (last-name (emchat-v8-fetch-lnts tlv1))
			    (email (emchat-v8-fetch-lnts tlv1))
			    (auth (emchat-v8-fetch-byte tlv1))
			    (status (emchat-v8-fetch-word-le tlv1))
			    (gender (emchat-v8-get-gender
				     (emchat-v8-fetch-byte tlv1)))
			    (age (emchat-v8-fetch-word-le tlv1))
			    (missed (emchat-v8-fetch-word-le tlv1)))
			(emchat-v8-debug "FOUND USER: UIN=%d, Nick=%s, missed=%d" uin nick missed)
			(emchat-v8-ctx-run-incoming-handler ectx 'search-found-last
			  :uin uin :nick nick
			  :first-name first-name :last-name last-name
			  :email email :auth auth :status status :gender gender
			  :age age :missed missed)))
		     )))))
    ))

(defun emchat-v8-snac-srv-toicqerr (ectx data &optional flags)
  "SRV_TOICQERR, SNAC(0x15, 0x01)"
  (let ((ecode (emchat-v8-fetch-word data))
	(tlv1 (emchat-v8-fetch-single-tlv data)))
    (error "Malformed TOICQ request" ecode tlv1)))

(defun emchat-v8-snac-srv-set-interval (ectx data &optional flags)
  "SRV_SETINTERVAL, SNAC(0x0B, 0x02)"
  (let ((ival (emchat-v8-fetch-word data)))
    (emchat-v8-debug "User info interval: %S" ival)
    (emchat-v8-ctx-put-prop ectx 'user-info-interval ival)))

(defun emchat-v8-snac-srv-icbm-err (ectx data &optional flags)
  "SRV_ICBMERR"
  (let* ((ival (emchat-v8-fetch-word data))
	 (err (cond ((= ival 4) "User offline")
		    ((= ival 9) "Client does not understand type-2 mesasges")
		    ((= ival 14) "Packet was malformed")
		    (t "Unknown"))))
    (emchat-v8-debug "ICBMERR: %d/%s" ival err)))

(defun emchat-v8-snac-srv-reply-icbm (ectx data &optional flags)
  "SRV_REPLICBM, SNAC(#x04, #x05)"
  )

(defun emchat-v8-snac-srv-onlinexinfo (ectx data &optional flags)
  "SRV_ONLINExINFO, SNAC(0x01, 0x0F)."
  ;; More info from the server about client IP and stuff.  Ignoring
  ;; for now, but possibly be useful if we ever do file transfers and
  ;; such.
  )

(defun emchat-v8-snac-srv-replyinfo (ectx data &optional flags)
  "SRV_REPLYINFO, SNAC(0x01, 0x15)"
  ;; This packet just tells us what the ICQ server thinks our IP is.
  ;; I don't really see the point in that, so ignoring this.
  )

(defun emchat-v8-snac-srv-reply-buddy (ectx data &optional flags)
  "SRV_REPLYBUDDY, SNAC(0x03, 0x03)"
  (let* ((tlvs (emchat-v8-fetch-tlvs data))
	 (muaicl (emchat-v8-tlv-get tlvs 1)) ;maximum uins allowed in contact list
	 (mpchuu (emchat-v8-tlv-get tlvs 2))) ;maximum number of people that can have our uin in contact list
    (when muaicl
      (emchat-v8-ctx-put-prop ectx 'maximum-uins (emchat-v8-tlv-num muaicl)))
    (when mpchuu
      (emchat-v8-ctx-put-prop ectx 'maximum-people (emchat-v8-tlv-num mpchuu)))))

(defun emchat-v8-snac-srv-cont-refused (ectx data &optional flags)
  "SRV_REFUSE, SNAC(0x03, 0x0A)."
  (let ((buin (emchat-v8-fetch-buin data)))
    (emchat-log-info "CANNOT watch status updates for: %s"
		     (emchat-uin-alias (emchat-stringular-uin buin)))
    (emchat-v8-debug "SRV_REFUSE: Cannot watch status of uin=%d" buin)))

(defun emchat-v8-snac-srv-user-online (ectx data &optional flags)
  "SRV_USERONLINE, SNAC(0x03, 0x0B)."
  (let* ((buin (emchat-v8-fetch-buin data))
	 (tlvs (progn
		 (emchat-v8-fetch-data data 4) ; skip
		 (emchat-v8-fetch-tlvs data)))
	 (ts (emchat-v8-tlv-get tlvs 6))
	 ;; XXX use only last 2 bytes of status
	 (new-status (and ts (emchat-v8-find-status
			      (progn
				(emchat-v8-fetch-word (emchat-v8-tlv-val ts))
				(emchat-v8-fetch-word (emchat-v8-tlv-val ts)))))))
    (emchat-v8-debug "SRV_USERONLINE: uin=%d status=%S/%S"
		   buin ts new-status)
    (when new-status
      (emchat-v8-ctx-run-incoming-handler ectx 'status-update
	buin new-status))))

(defun emchat-v8-snac-srv-user-offline (ectx data &optional flags)
  "SRV_USEROFFLINE, SNAC(0x03, 0x0C)"
  (let ((buin (emchat-v8-fetch-buin data)))
    (emchat-v8-debug "SRV_USEROFFLINE: uin=%d" buin)
    (emchat-v8-ctx-run-incoming-handler ectx 'status-update
      buin 'offline)))

(defun emchat-v8-snac-srv-reply-location (ectx data &optional flags)
  "SRV_REPLYLOCATION, SNAC(0x02, 0x03)."
  ;; TODO: -write me
  (let* ((tlvs (emchat-v8-fetch-tlvs data))
	 (mcap (emchat-v8-tlv-get tlvs 2)))
    (when mcap
      ;; Set maximum capabilities
      (emchat-v8-debug "max capabilities: %S" (emchat-v8-tlv-num mcap))
      (emchat-v8-ctx-put-prop ectx 'maximum-capabilities (emchat-v8-tlv-num mcap)))))

(defun emchat-v8-snac-srv-families (ectx data &optional flags)
  "SRV_FAMILIES, SNAC(0x01, 0x03)."
  (let (srv-families)
    (while (> (length (aref data 0)) 0)
      (setq srv-families
	    (cons (emchat-v8-fetch-word data) srv-families)))
    (emchat-v8-ctx-put-prop ectx 'server-families srv-families))

  (emchat-v8-debug "SRV-FAMILIES %S" (emchat-v8-ctx-get-prop ectx 'server-families))
  (emchat-v8-snac-cli-families ectx))

(defun emchat-v8-snac-srv-families2 (ectx data &optional flags)
  "SRV_FAMILIES2, SNAC(0x01, 0x12)."
  (emchat-v8-debug "SRV-FAMILIES2 %S" data)
  )

(defun emchat-v8-snac-bart-request (ectx data &optional flags)
  "BART_REQUEST, SNAC(0x01, 0x21)."
  (let* ((type (emchat-v8-fetch-word data))
	 (flag (emchat-v8-fetch-byte data))
	 (len_opaque (emchat-v8-fetch-byte data))
	 (opaque (emchat-v8-fetch-data data len_opaque)))
    (emchat-v8-debug "[BART]...
Type: %S
Flag: %S
len:  %S
opaque: %S"
		     type flag len_opaque opaque)))

(defun emchat-v8-snac-srv-rates (ectx data &optional flags)
  "SRV_RATES, SNAC(0x01, 0x07)."
  (let ((nr (emchat-v8-parse-message
	     data
	     (list
	      [2 emchat-v8-length-1]
	      [emchat-v8-length-1
	       ([2 number]		; Rate class ID
		[4 number]		; Window size
		[4 number]		; Clear level
		[4 number]		; Alert level
		[4 number]		; Limit level
		[4 number]		; Disconnect level
		[4 number]		; Current level
		[4 number]		; Max level
		[4 number]		; last time
		[1 number])]		; current state
	      [emchat-v8-length-1
	       ([2 number]		; Rate class ID
		[2 emchat-v8-length-2]	; number of pairs in group
		[emchat-v8-length-2 ([4 number])])]))))

    (emchat-v8-debug "SRV_RATES: %S" nr)

    ;; Send a reply
    (when (cadr nr)
      (emchat-v8-send ectx
	(emchat-v8-pack-snac ectx '(#x01 . #x08)
	  (apply 'concat
		 (mapcar 'emchat-v8-pack-word (mapcar 'car (nth 1 nr)))))))
    ))

(defun emchat-v8-snac-srv-rate-change (ectx data &optional flags)
  "SRV_RATE_CHANGE, SNAC(#x01, #x0A)."
  (let ((code (emchat-v8-fetch-word data))
	(rateid (emchat-v8-fetch-word data))
	(winsize (emchat-v8-fetch-dword data))
	(clear (emchat-v8-fetch-dword data))
	(alert (emchat-v8-fetch-dword data))
	(limit (emchat-v8-fetch-dword data))
	(discon (emchat-v8-fetch-dword data))
	(currl (emchat-v8-fetch-dword data))
	(max (emchat-v8-fetch-dword data))
	(ltime (emchat-v8-fetch-dword data))
	(currs (emchat-v8-fetch-byte data)))
    (emchat-v8-debug
     "[SRV_RATE_EXCEEDED]:
    Message Code: %d
   Rate Class ID: %d
     Window Size: %d
     Clear Level: %d
     Alert Level: %d
     Limit Level: %d
Disconnect Level: %d
   Current Level: %d
       Max Level: %d
       Last Time: %d
   Current State: %d"
     code rateid winsize clear alert limit discon currl max ltime currs)
    (ecase code
      (1 #'ignore)
      (2 (emchat-log-info "You've reached the rate limit: %d
Chill out, take a break, have a coffee." alert))
      (3 (emchat-log-error "Rate limit passed: %d
Hope you get to read this before you get disconnected" discon))
      (4 #'ignore))))

(defun emchat-v8-snac-srv-pause (ectx data &optional flags)
  "SRV_SERVERPAUSE, SNAC(1, B)."
  ;; Send CLI_ACKSERVERPAUSE snac
  (emchat-v8-send ectx
    (emchat-v8-pack-snac ectx '(#x09 . #x02)
      (apply 'concat (mapcar #'(lambda (fam)
				 (emchat-v8-pack-word (car fam)))
			     emchat-v8-snacv-list))))

  ;; Mark ECTX as in pause, until SRV_MIGRATIONREQ is received
  (setf (emchat-v8-ctx-pause-p ectx) t))

(defun emchat-v8-snac-srv-migrate (ectx data &optional flags)
  "SRV_MIGRATIONREQ, SNAC(1, 12)."
  (let* ((tlvs (progn
		 (emchat-v8-fetch-word data)
		 (emchat-v8-fetch-tlvs data)))
	 (tlv5 (emchat-v8-tlv-get tlvs 5))
	 (tlv6 (emchat-v8-tlv-get tlvs 6)))

    ;; Close current connection
    (emchat-v8-close ectx)

    (let* ((nsrv (split-string (emchat-v8-tlv-str tlv5) ":"))
	   (addr (nth 0 nsrv))
	   (port (string-to-int (nth 1 nsrv))))
      (emchat-v8-debug "Migrating to new server: %s:%d" addr port)
      ;; Store cookie and reconnect
      (emchat-v8-ctx-put-prop ectx 'cli-cookie (emchat-v8-tlv-str tlv6))
      (emchat-v8-connect ectx addr port))))

(defun emchat-v8-snac-srv-motd (ectx data &optional flags)
  "SRV_MOTD, SNAC(1, 13)."
  (emchat-v8-debug "SRV_MOTD")

  (emchat-v8-snac-cli-ratesrequest ectx)
  (emchat-v8-snac-cli-reqinfo ectx)

  ;; TODO: roster
  (emchat-v8-snac-cli-reqlocation ectx)
  (emchat-v8-snac-cli-reqbuddy ectx)
  (emchat-v8-snac-cli-reqicbm ectx)
  (emchat-v8-snac-cli-reqbos ectx)

  (emchat-v8-ctx-run-incoming-handler ectx 'connected))

(defconst emchat-v8-msg-types
  '((0 . automatic)                     ; An automatic message
    (1 . normal)                        ; A plain normal message
    (2 . chat-request)                  ; A chat request
    (3 . file-transfer-request)         ; A file transfer request
    (4 . url)				; An URL message. The message consists of the description and the url
    (6 . auth-request)			; An authorization request
    (7 . auth-reject)			; An authorization reject message
    (8 .  auth-accept)			; An authorization accept message
    (12 . added)			; You were added to the sender's contact list
    (13 . web-pager)                    ; A message sent through www.icq.com's web pager
    (14 . email-pager)			; A message sent through the email pager ([uin]@pager.icq.com)
    (19 . contact-list)			; A contact list message
    (26 . extended)			; An extended message. The packet will contain more data
    (1000 . get-away)			; A message requesting the "away" auto message
    (1001 . get-occ)			; A message requesting the "occupied" auto message
    (1002 . get-na)			; A message requesting the "not available" auto message
    (1003 . get-dnd)			; A message requesting the "do not disturb" auto message
    (1004 . get-ffc)			; A message requesting the "free for chat" auto message
    )
  "List of message types.")

(defconst emchat-v8-msg-mass-flag #x8000
  "If set, message was sent to several recipients.")

(defun emchat-v8-snac-srv-recv-msg (ectx data &optional flags)
  "SRV_RECVMSG, SNAC(4, 7)."
  (emchat-v8-debug "data here: %S" data)
  (let ((midtime (emchat-v8-fetch-time data))
	(midrand (emchat-v8-fetch-dword data))
	(type (emchat-v8-fetch-word data))
	(uin (emchat-v8-fetch-buin data))
	(warn (emchat-v8-fetch-word data))
	(count (emchat-v8-fetch-word data))
	(tlvs (emchat-v8-fetch-tlvs data)))

    ;; Fetch tlvs
    (emchat-v8-debug "Got message from %S, mt/mr/t/w/c=%S/%S/%S/%S/%S, tlvs=%S"
		   uin midtime midrand type warn count tlvs)

    (cond ((= type 1)
	   ;; Normal message
	   (let* ((tlvals (emchat-v8-fetch-tlvs
			   (emchat-v8-tlv-val (emchat-v8-tlv-get tlvs 2))))
		  (caps (emchat-v8-fetch-byte
			 (emchat-v8-tlv-val (emchat-v8-tlv-get tlvals 1281))))
		  (mtlval (emchat-v8-tlv-val
			   (emchat-v8-tlv-get tlvals 257)))
		  (menc (emchat-v8-fetch-word mtlval))
		  (msg (progn
			 (emchat-v8-fetch-word mtlval) ; skip
			 (emchat-v8-fetch-string mtlval))))

	     (emchat-v8-debug "Normal message: caps/menc=%d/%d, msg='%s'"
			    caps menc msg)
	     (emchat-v8-ctx-run-incoming-handler ectx 'instant-message
	       :uin uin :msg msg :msg-type 'normal)))

	  ((= type 2)
	   ;; Advanced message
	   ;; TODO: -write me
	   (emchat-v8-debug "MSG[%d]: here" type))

	  ((= type 4)
	   ;; Server message
	   (let* ((tlv5 (emchat-v8-tlv-val (emchat-v8-tlv-get (nreverse tlvs) 5)))
		  (uin (truncate (emchat-v8-fetch-uin tlv5)))
		  (mtype (emchat-v8-fetch-word-le tlv5))
		  (msg (emchat-v8-fetch-lnts tlv5))
		  msg-type)
	     ;; Adjust MTYPE to remove flag fields (MSGF_MASS)
	     (setq mtype (logand mtype (lognot emchat-v8-msg-mass-flag)))
	     (setq msg-type (cdr (assoc mtype emchat-v8-msg-types)))

	     (emchat-v8-ctx-run-incoming-handler ectx 'instant-message
		 :uin uin :msg msg :msg-type msg-type)))

	  (t (error "Unknown message type")))))

(defun emchat-v8-snac-srv-missed-icbm (ectx data &optional flags)
  "SRV_MISSED_ICBM, SNAC(4, 10)."
  (let* ((buin (progn
		 (emchat-v8-fetch-word data) ; skip channel
		 (emchat-v8-fetch-buin data)))
	 (mmsgs (progn
		  (emchat-v8-fetch-word data) ; skip warn
		  (dotimes (n (emchat-v8-fetch-word data)) ; skip tlvs
		    (emchat-v8-fetch-single-tlv data))
		  (emchat-v8-fetch-word data)))
	 (rn (emchat-v8-fetch-word data))
	 (reasons '((0 . "Invalid message")
		    (1 . "Message too large")
		    (2 . "Message rate limit exceeded")
		    (3 . "Sender too evil (sender warn level > your max_msg_sevil)")
		    (4 . "You are too evil (sender max_msg_revil > your warn level)")))
	 (reason (or (cdr (assq rn reasons)) "Unknown reason")))
    (emchat-v8-ctx-run-incoming-handler ectx 'missed-message
      :uin buin :missed-messages mmsgs :reason reason)))

(defalias 'emchat-v8-snac-srv-ack-msg #'ignore)

(defun emchat-v8-snac-srv-contact-err (ectx data &optional flags)
  "SRV_CONTACTERR, SNAC(3, #x01)."
  (let* ((ercode (emchat-v8-fetch-word data))
	 (errmsg (cond ((= ercode #x0E) "Empty UIN list sent")
		       ((= ercode #x14) "Non existant UIN")
		       ((= ercode #x15) "Contact list full"))))
    (emchat-v8-debug "[SRV_CONTACTERR]: Code=%d (0x%1$02X), Msg=%s" ercode errmsg)
    (emchat-v8-ctx-run-incoming-handler ectx 'srv-contacterr :reason errmsg)))

(defun emchat-v8-snac-srv-err (ectx data &optional flags)
  "SRV_GEN_ERR, SNAC(1, #x01)."
  (let* ((ercode (emchat-v8-fetch-word data))
	 (errmsg (cond ((= ercode #x01) "Invalid SNAC header.")
		       ((= ercode #x02) "Server rate limit exceeded")
		       ((= ercode #x03) "Client rate limit exceeded")
		       ((= ercode #x04) "Recipient is not logged in")
		       ((= ercode #x05) "Requested service unavailable")
		       ((= ercode #x06) "Requested service not defined")
		       ((= ercode #x07) "You sent obsolete SNAC")
		       ((= ercode #x08) "Not supported by server")
		       ((= ercode #x09) "Not supported by client")
		       ((= ercode #x0A) "Refused by client")
		       ((= ercode #x0B) "Reply too big")
		       ((= ercode #x0C) "Responses lost")
		       ((= ercode #x0D) "Request denied")
		       ((= ercode #x0E) "Incorrect SNAC format")
		       ((= ercode #x0F) "Insufficient rights")
		       ((= ercode #x10) "In local permit/deny (recipient blocked)")
		       ((= ercode #x11) "Sender too evil")
		       ((= ercode #x12) "Receiver too evil")
		       ((= ercode #x13) "User temporarily unavailable")
		       ((= ercode #x14) "No match")
		       ((= ercode #x15) "List overflow")
		       ((= ercode #x16) "Request ambiguous")
		       ((= ercode #x17) "Server queue full")
		       ((= ercode #x18) "Not while on AOL")
		       (t "Unknown error"))))
    (emchat-v8-debug "SRV_GEN_ERR: %s" errmsg)
    (emchat-v8-ctx-run-incoming-handler ectx 'srv-error :reason errmsg)))

(defun emchat-v8-snac-srv-reply-bos (ectx data &optional flags)
  "SRV_REPLYBOS, SNAC(9, 3)."
  (emchat-v8-snac-cli-setuserinfo ectx)
  (emchat-v8-snac-cli-setstatus
   ectx (emchat-v8-ctx-get-prop ectx 'initial-status))
  (emchat-v8-snac-cli-ready ectx)
  ;; This is the now unsupported local-only type of contact list.
  ;; (emchat-v8-snac-cli-add-contact ectx)
  (emchat-v8-snac-cli-reqofflinemsgs ectx))

(defun emchat-v8-snac-srv-srv-ack-msg (ectx data &optional flags)
  "SRV_SRVACKMSG, SNAC (4 . #x0C)"
  (let ((type (progn
		(emchat-v8-fetch-data data 8) ; skip
		(emchat-v8-fetch-word data)))
	(buin (emchat-v8-fetch-buin data)))
    (emchat-v8-debug "SRVACKMSG: type=%d uin=%d" type buin)
    ))

;;;; CLI

;;; Family #x13 (SSI -- Server Side Info)

(defun emchat-v8-snac-cli-ssi-right-request (ectx)
  "Request rights/limitations for SSI. SNAC(#x13, #x02)."
  (emchat-v8-send ectx
    (emchat-v8-pack-snac ectx '(#x13 . #x02))))

(defun emchat-v8-snac-cli-ssi-request (ectx)
  "Request server side info.  SNAC(#x13, #x04)."
  (emchat-v8-send ectx
    (emchat-v8-pack-snac ectx '(#x13 . #x04))))

(defun emchat-v8-snac-cli-ssi-checkout (ectx)
  "Checkout SSI list if local copy is out of date.  SNAC(#x13, #x05)."
  (let ((num (emchat-world-ssi-count))
	(modt (emchat-world-ssi-mod-time)))
    (emchat-v8-send ectx
      (emchat-v8-pack-snac ectx '(#x13 . #x05)
	(emchat-v8-pack-time modt)
	(emchat-v8-pack-word num)))))

(defun emchat-v8-snac-cli-ssi-activate (ectx)
  "Activate SSI.  SNAC(#x13, #x07)."
  (emchat-v8-send ectx
    (emchat-v8-pack-snac ectx '(#x13 . #x07))))

;;; FIXME:  Write me
(defalias 'emchat-v8-snac-cli-ssi-add #'ignore)
;; (defun emchat-v8-snac-cli-ssi-add (ectx uin grp id nick)
;;   "Add contacts to SSI.  SNAC(#x13, #x08)."
;;   (emchat-v8-send ectx
;;     (emchat-v8-pack-snac ectx '(#x13 . #x08)
;;       (emchat-v8-pack-bstr uin)
;;       (emchat-v8-pack-word grp)
;;       (emchat-v8-pack-word id)
;;       (emchat-v8-pack-word #x0000)
;;       (emchat-v8-pack-tlv #x0131 nick))))

;;; FIXME: Write me
(defalias 'emchat-v8-snac-cli-ssi-update #'ignore)
;; (defun emchat-v8-snac-cli-ssi-update (ectx &rest entries)
;;   "Update existing SSI entries.  SNAC(#x13, #x09)."
;;   (emchat-v8-debug "CLI_SSI_UPDATE: somebody impliment me"))

;;; FIXME: Write me
(defalias 'emchat-v8-snac-cli-ssi-delete #'ignore)
;; (defun emchat-v8-snac-cli-ssi-delete (ectx &rest entries)
;;   "Delete existing SSI entries.  SNAC(#x13, #x0A)."
;;   (emchat-v8-debug "CLI_SSI_DELETE: somebody impliment me"))
(defalias 'emchat-v8-snac-cli-ssi-edit-begin #'ignore)
;; (defun emchat-v8-snac-cli-ssi-edit-begin (ectx &rest args)
;;   "Begin SSI edits.  SNAC(#x13, #x11)."
;;   (emchat-v8-send ectx
;;     (emchat-v8-pack-snac ectx '(#x13 . #x11))))
(defalias 'emchat-v8-snac-cli-ssi-edit-end #'ignore)
;; (defun emchat-v8-snac-cli-ssi-edit-end (ectx &rest args)
;;   "End SSI edits.  SNAC(#x13, #x12)."
;;   (emchat-v8-send ectx
;;     (emchat-v8-pack-snac ectx '(#x13 . #x12))))

(defun emchat-v8-snac-cli-ssi-future-auth-grant (ectx uin reason)
  "Give future auth grant.  SNAC(#x13, #x14)."
  (emchat-v8-send ectx
    (emchat-v8-pack-snac ectx '(#x13 . #x14)
      (emchat-v8-pack-byte (length (number-to-string uin)))
      (emchat-v8-pack-buin uin)
      (emchat-v8-pack-dword 0))))

(defun emchat-v8-snac-cli-ssi-del-yourself (ectx uin)
  "Delete yourself from someone's contact list.  SNAC(#x13, #x16)."
  (emchat-v8-send ectx
    (emchat-v8-pack-snac ectx '(#x13 . #x16)
      (emchat-v8-pack-buin uin))))

(defun emchat-v8-snac-cli-ssi-send-auth-request (ectx uin msg)
  "Send auth request.  SNAC(#x13, #x18)."
  (emchat-v8-send ectx
    (emchat-v8-pack-snac ectx '(#x13 . #x18)
      (emchat-v8-pack-buin uin)
      (emchat-v8-pack-bstr msg))))

(defun emchat-v8-snac-cli-ssi-auth-reply (ectx uin reply &optional reason)
  "Reply to auth-request.  SNAC(#x13, #x1A)."
  (emchat-v8-send ectx
    (emchat-v8-pack-snac ectx '(#x13 . #x1A)
      (emchat-v8-pack-buin uin)
      (emchat-v8-pack-byte reply)
      (emchat-v8-pack-bstr reason)
      (emchat-v8-pack-word 0))))


(defun emchat-v8-snac-cli-setuserinfo (ectx)
  "Send userinfo."
  )

(defconst emchat-v8-status-alist
  `((offline . ,(* 256.0 256 256 256))
    (invisible . #x100)			;user is invisible
    (dnd . #x02)		        ;user does not want to be disturbed
    (occupied . #x10)			;user is occupied
    (na . #x04)				;user not available
    (away . #x01)			;user is away
    (ffc . #x20)			;user is free for chat
    (online . #x00)			;user is online
    (set-invisible . #x100)
    (set-dnd . #x13)
    (set-occupied . #x11)
    (set-na . #x05)
    (set-away . #x01)
    (set-ffc . #x20)
    (web-aware . #x010000)
    (allow-ip . #x020000)
    (birthday . #x80000)
    (dcauth . #x10000000)
    (dccontact . #x20000000))
  "Alist of states.")

(defun emchat-v8-find-status (num)
  "Find statu by NUM."
  (setq num (logand num #x0000ffff))
  (let ((statuses (cdr emchat-v8-status-alist)))
    (while (and statuses
		(zerop (logand num (cdar statuses))))
      (setq statuses (cdr statuses)))
    (or (caar statuses) 'online)))

(defun emchat-v8-snac-cli-setstatus (ectx &optional status)
  "Set status, SNAC(1, 30)."
  (setq status (apply 'logior
		      (mapcar #'(lambda (st)
				  (cdr (assoc st emchat-v8-status-alist)))
			      status)))

  (emchat-v8-send ectx
    (emchat-v8-pack-snac ectx '(#x01 . #x1E)
      (emchat-v8-pack-tlv 6
	(emchat-v8-pack-dword status))
      (emchat-v8-pack-tlv 8
	(emchat-v8-pack-word 0)))))

(defun emchat-v8-snac-cli-add-contact (ectx &optional contacts)
  "CLI_ADDCONTACT, SNAC(3,4)
Send either CONTACTS or ECTX's contact list."
  (unless contacts
    (setq contacts (emchat-v8-ctx-get-prop ectx 'contacts)))
  (emchat-v8-debug "Adding contacts: %S" contacts)
  (when contacts
    (emchat-v8-send ectx
      (emchat-v8-pack-snac ectx '(#x03 . #x04)
	(emchat-v8-pack-uinlist contacts)))))

(defun emchat-v8-snac-cli-addvisible (ectx &optional uin-list)
  "CLI_ADDVISIBLE SNAC(9, 5)."
  (unless uin-list
    (setq uin-list (emchat-v8-ctx-get-prop ectx 'visible-list)))
  (emchat-v8-debug "Adding to visible: %S" uin-list)
  (when uin-list
    (emchat-v8-send ectx
      (emchat-v8-pack-snac ectx '(#x09 . #x05)
	(emchat-v8-pack-uinlist uin-list)))))

(defun emchat-v8-snac-cli-remvisible (ectx uin-list)
  "CLI_ADDVISIBLE SNAC(9, 6)."
  (emchat-v8-debug "Removing from visible: %S" uin-list)
  (when uin-list
    (emchat-v8-send ectx
      (emchat-v8-pack-snac ectx '(#x09 . #x06)
	(emchat-v8-pack-uinlist uin-list)))))

(defun emchat-v8-snac-cli-addinvisible (ectx &optional uin-list)
  "CLI_ADDVISIBLE SNAC(9, 7)."
  (unless uin-list
    (setq uin-list (emchat-v8-ctx-get-prop ectx 'invisible-list)))
  (emchat-v8-debug "Adding to invisible: %S" uin-list)
  (when uin-list
    (emchat-v8-send ectx
      (emchat-v8-pack-snac ectx '(#x09 . #x07)
	(emchat-v8-pack-uinlist uin-list)))))

(defun emchat-v8-snac-cli-reminvisible (ectx uin-list)
  "CLI_REMINVISIBLE SNAC(9, 8)."
  (emchat-v8-debug "Removing from invisible: %S" uin-list)
  (when uin-list
    (emchat-v8-send ectx
      (emchat-v8-pack-snac ectx '(#x09 . #x08)
	(emchat-v8-pack-uinlist uin-list)))))

(defun emchat-v8-snac-cli-keepalive (ectx)
  "Send keep-alive flap."
  (emchat-v8-send ectx
    (emchat-v8-pack-flap ectx 5)))

(defun emchat-v8-pack-meta-snac (ectx sub &optional type &rest data)
  "Create meta snac."
  (setq data (apply 'concat data))

  (let ((seq3 (emchat-v8-ctx-get-prop ectx 'our-seq3))
	temp-data)
    (if seq3
	(emchat-v8-ctx-put-prop ectx 'our-seq3 (% (1+ seq3) 32767))
      (emchat-v8-ctx-put-prop ectx 'our-seq3 2))

    (setq temp-data
	  (concat
	   (emchat-v8-pack-uin (plist-get (emchat-v8-ctx-userinfo ectx) 'uin))
	   (emchat-v8-number->string-swap 2 sub)
	   (emchat-v8-number->string-swap 2 (emchat-v8-ctx-get-prop ectx 'our-seq3))
	   (when type
	     (emchat-v8-number->string-swap 2 type))
	   data))

    (emchat-v8-pack-snac ectx '(#x15 . #x02)
      (emchat-v8-pack-tlv 1
	(emchat-v8-number->string-swap 2 (length temp-data))
	temp-data))))

(defun emchat-v8-snac-cli-reqofflinemsgs (ectx)
  "CLI_REQOFFLINEMSGS, SNAC(15,2)."
  (emchat-v8-send ectx
    (emchat-v8-pack-meta-snac ectx 60)))

(defun emchat-v8-snac-cli-metareqinfo (ectx uin)
  "CLI_METAREQINFO, SNAC(15,2)/2000/1232."
  (emchat-v8-send ectx
    (emchat-v8-pack-meta-snac ectx 2000 1232
       (emchat-v8-pack-uin uin)))

  ;; Save pair SEQ/UIN for later UIN extraction
  (emchat-v8-ctx-put-prop ectx 'about-query-seq-uin
    (cons (cons (emchat-v8-ctx-get-prop ectx 'our-seq3) uin)
	  (emchat-v8-ctx-get-prop ectx 'about-query-seq-uin))))

(defun emchat-v8-snac-cli-searchbyuin (ectx uin)
  "CLI_SEARCHBYUIN, SNAC(15,2)/2000/1385."
  (emchat-v8-send ectx
    (emchat-v8-pack-meta-snac ectx 2000 1385
       (emchat-v8-pack-word-le 310)       ; Search key (310 == uin)
       (emchat-v8-pack-lnts (emchat-v8-pack-uin uin)))))

(defun emchat-v8-snac-cli-searchbyemail (ectx email)
  "CLI_SEARCHBYEMAIL, SNAC(15,2)/2000/1395."
  (emchat-v8-send ectx
    (emchat-v8-pack-meta-snac
     ectx 2000 1395
     (emchat-v8-pack-word-le 350)		; Search key (350 == email)
     (emchat-v8-pack-llnts email))))

(defun emchat-v8-snac-cli-searchbypersinf (ectx first last nick email online)
  "CLI_SEARCHBYPERSINF, SNAC(15,2)/2000/1375."
  (emchat-v8-send ectx
    (emchat-v8-pack-meta-snac
     ectx 2000 1375
     (emchat-v8-pack-word-le 320)	; Search key (320 == first name)
     (emchat-v8-pack-llnts first)
     (emchat-v8-pack-word-le 330)	; Search key (330 == last name)
     (emchat-v8-pack-llnts last)
     (emchat-v8-pack-word-le 340)	; Search key (340 == nick name)
     (emchat-v8-pack-llnts nick)
     (emchat-v8-pack-word-le 350)	; Search key (350 == email)
     (emchat-v8-pack-llnts email)
     (emchat-v8-pack-byte online))))

;;; FIXME: this doesn't impliment direct connect or something called
;;; "user kind"
(defun emchat-v8-snac-cli-metasetsecurity (ectx auth web)
  "CLI_METASETSECURITY. SNAC(#x15, #x02)/2000/1060."
  (emchat-v8-send ectx
    (emchat-v8-pack-meta-snac
     ectx 2000 1060
     (emchat-v8-pack-byte auth)		; req auth 0 = no; 1 = yes
     (emchat-v8-pack-byte web)		; webaware 0 = yes; 1 = no
     (emchat-v8-pack-byte #x02)		; dc hard coded to only with auth
     (emchat-v8-pack-byte 0))))		; user kind (no idea, set to zero)

(defun emchat-v8-snac-cli-ready (ectx)
  "CLI_READY, SNAC(1, 2)."
  (emchat-v8-send ectx
    (emchat-v8-pack-snac ectx '(#x01 . #x02)
      (apply 'concat
	     (mapcar #'(lambda (sn)
			 (concat
			  (emchat-v8-pack-word (car sn))
			  (emchat-v8-pack-word (cdr sn))
			  (if (= (car sn) 2)
			      (emchat-v8-pack-dword #x0101047B)
			    (emchat-v8-pack-dword #x0110047B))))
		     emchat-v8-snacv-list)))))

(defun emchat-v8-snac-cli-families (ectx)
  "CLI_FAMILIES, SNAC(1, 17)."
  (emchat-v8-send ectx
    (emchat-v8-pack-snac ectx '(#x01 . #x17)
      (apply 'concat
	     (mapcar #'(lambda (sn)
			 (concat
			  (emchat-v8-pack-word (car sn))
			  (emchat-v8-pack-word (cdr sn))))
		     emchat-v8-snacv-list)))))

(defun emchat-v8-snac-cli-ratesrequest (ectx)
  "CLI_RATESREQUEST, SNAC(1, 6)."
  (emchat-v8-send ectx
    (emchat-v8-pack-snac ectx '(#x01 . #x06))))

(defun emchat-v8-snac-cli-reqinfo (ectx)
  "CLI_REQINFO, SNAC(1, E)."
  (emchat-v8-send ectx
    (emchat-v8-pack-snac ectx '(#x01 . #x0E))))

(defun emchat-v8-snac-cli-reqlocation (ectx)
  "CLI_REQLOCATION, SNAC(2, 2)."
  (emchat-v8-send ectx
    (emchat-v8-pack-snac ectx '(#x02 . #x02))))

(defun emchat-v8-snac-cli-reqbuddy (ectx)
  "CLI_REQBUDDY, SNAC(3, 2)."
  (emchat-v8-send ectx
    (emchat-v8-pack-snac ectx '(#x03 . #x02))))

(defun emchat-v8-snac-cli-reqicbm (ectx)
  "CLI_REQICBM, SNAC(4, 4)."
  (emchat-v8-send ectx
    (emchat-v8-pack-snac ectx '(#x04 . #x04))))

(defun emchat-v8-snac-cli-sendmsg (ectx uin format msg)
  "CLI_SENDMSG, SNAC(4, 6).
UIN - Icq uin to send message to.
FORMAT - format id, 1 - simple, 2 - advanced, 4 - typed.
MSG - Message itself."
  (emchat-v8-send ectx
    (emchat-v8-pack-snac ectx '(#x04 . #x06)
      (emchat-v8-pack-dword 0)            ; random
      (emchat-v8-pack-dword 0)            ; random
      (emchat-v8-pack-word format)
      (emchat-v8-pack-buin uin)
      msg
      (emchat-v8-pack-tlv 6))))

(defconst emchat-v8-msg-encodings
  '((us-ascii . 0)
    (utf-16 . 2)
    (8bit . 3)))

(defun emchat-v8-send-simple-message (ectx uin message)
  "Send simple message."
  (emchat-v8-snac-cli-sendmsg ectx uin 1
    (emchat-v8-pack-tlv 2
      (emchat-v8-pack-tlv 1281 (emchat-v8-pack-byte 1))
      (emchat-v8-pack-tlv 257
	(emchat-v8-pack-word 3)
	(emchat-v8-pack-word (+ 4 (length message)))
	message))))

(defun emchat-v8-send-typed-message (ectx uin type message)
  "Send typed message."
  (emchat-v8-snac-cli-sendmsg ectx uin 4
    (emchat-v8-pack-tlv 5
      (emchat-v8-pack-uin (plist-get (emchat-v8-ctx-userinfo ectx) 'uin))
      (emchat-v8-pack-word-le
       (car (find type emchat-v8-msg-types :key 'cdr :test 'eq)))
      (emchat-v8-pack-lnts message))))

(defun emchat-v8-send-advanced-message (ectx uin &rest add-arguments-here-for-advanced-message)
  ;; Advanced message:
  ;; TLV(5):
  ;;    2 ACKTYPE :  0 - normal, 1 - abort, 2 - file ack
  ;;    4 TIME : copy
  ;;    4 ID : random id
  ;;  TLV(11) UNKNOWN : Only if ACTYPE is 1
  ;;   CAP
  ;;  TLV(10) ACKTYPE2 : 1 - normal, 2 - file ack
  ;;  TLV(5) PORT
  ;;  TLV(3) IP
  ;;  TLV15  UNKNOWN
  ;;  TLV(1001) MESSAGE
;  (emchat-v8-snac-cli-sendmsg ectx uin 2)
  (emchat-v8-debug "TODO: `emchat-v8-send-advanced-message'")
  )

(defun emchat-v8-send-server-message (ectx uin &rest add-arguments-here-for-server-message)
  ;; Server message:
  ;; TLV(5)
  ;;   4  UIN  : sender uin
  ;;   2  MSGTYPE : message type
  ;; LNTS MSG : message with unspecified encoding
;  (emchat-v8-snac-cli-sendmsg ectx uin 3)
  (emchat-v8-debug "TODO: `emchat-v8-send-advanced-message'")
  )

(defun emchat-v8-snac-cli-reqbos (ectx)
  "CLI_REQBOS, SNAC(9, 2)."
  (emchat-v8-send ectx
    (emchat-v8-pack-snac ectx '(#x09 . #x02))))

(defun emchat-v8-handle-flap (ectx-flap)
  (let* ((ectx (car ectx-flap))
	 (flap (cdr ectx-flap))
	 (cid (emchat-v8-flap-cid flap)))
    (cond ((= cid emchat-v8-FLAP-HELLO)
	   ;; XXX Login channel
	   (emchat-v8-debug "Got HELLO cid")
	   (let ((cmd (emchat-v8-fetch-dword (emchat-v8-flap-data flap)))
		 (cli-cook (emchat-v8-ctx-get-prop ectx 'cli-cookie)))
	     (cond ((= cmd 1)
		    (if cli-cook
			;; Already logged in, send Cookie
			(emchat-v8-send ectx
			  (emchat-v8-pack-flap ectx emchat-v8-FLAP-HELLO
			    (emchat-v8-pack-dword 1)
			    (emchat-v8-pack-tlv 6 cli-cook)))

		      ;; Not logged in yet
		      (emchat-v8-login ectx)))

		   (t (error (format "Unknown FLAP CMD(%) in login channel." cmd))))
	     ))

	  ((= cid emchat-v8-FLAP-SNAC)
	   ;; SNAC channel
	   ;; TODO:
	   ;;  - Write me
	   (let ((snac (emchat-v8-fetch-snac (emchat-v8-flap-data flap)))
		 (snacs emchat-v8-snac-list))
	     (while (and snacs
			 (or (not (= (emchat-v8-snac-family snac)
				     (car (caar snacs))))
			     (not (= (emchat-v8-snac-subtype snac)
				     (cdr (caar snacs))))))
	       (setq snacs (cdr snacs)))

	     (emchat-v8-debug "SNAC %S (%S)" snac (when (car snacs) (nth 1 (car snacs))))

	     (unless snacs
	       (error 'invalid-argument "Unknown SNAC" snac))

	     ;; Call snac handler
	     (funcall (nth 2 (car snacs))
		      ectx		             ; context
		      (emchat-v8-flap-data flap)     ; data
		      (emchat-v8-snac-flags snac)))) ; snac flags

	  ((= cid emchat-v8-FLAP-ERRORS)
	   ;; ERRORS channel
	   (emchat-v8-debug "ERROR")
	   )

	  ((= cid emchat-v8-FLAP-LOGOFF)
	   ;; LOGOFF channel
	   (let* ((tlvs (emchat-v8-fetch-tlvs (emchat-v8-flap-data flap)))
		  (tlv (emchat-v8-tlv-get tlvs 5)))

	     (emchat-v8-debug "LOGOFF tlv_t5=%S" tlv)
	     ;; Close connection
	     (emchat-v8-close ectx)

	     (if (not tlv)
		 ;; Another user with same UIN
		 (emchat-v8-ctx-run-incoming-handler ectx 'logoff)

	       ;; Redirected
	       (let* ((nsrv (split-string (emchat-v8-tlv-str tlv) ":"))
		      (addr (nth 0 nsrv))
		      (port (string-to-int (nth 1 nsrv))))
		 (emchat-v8-debug "Redirected to new server: %s:%d" addr port)
		 ;; Store cookie and reconnect
		 (emchat-v8-ctx-put-prop ectx 'cli-cookie
		   (emchat-v8-tlv-str (emchat-v8-tlv-get tlvs 6)))
		 (emchat-v8-connect ectx addr port)))))

	  ((= cid emchat-v8-FLAP-PING)
	   ;; PING channel
	   (emchat-v8-debug "PING")
	   )

	  (t (error (format "Unknown FLAP cid(%d)." cid))))
    ))

(defun emchat-v8-fetch-handle-flaps (ectx)
  (while (> (length (emchat-v8-ctx-incoming-buffer ectx)) 0)
    ;; Do not do anything in filter function
    (emchat-v8-handle-flap (cons ectx (emchat-v8-fetch-flap ectx)))))


(defun emchat-v8-number->string (size val)
  "Convert number value VAL to string of SIZE."
  (cond ((= size 1)
	 (setq val (truncate val))
	 (char-to-string val))
	((= size 2)
	 (setq val (truncate val))
	 (concat (char-to-string (ash (mod val 65536) -8))
		 (char-to-string (logand val 255))
		 ))
	((= size 4)
	 ;; XXX remove this
	 (setq val (truncate val))
	 (mapconcat 'identity
		    (nreverse
		     (list (char-to-string (int-to-char (% val 256)))
			   (char-to-string (int-to-char (% (setq val (/ val 256)) 256)))
			   (char-to-string (int-to-char (% (setq val (/ val 256)) 256)))
			   (char-to-string (int-to-char (% (setq val (/ val 256)) 256)))
			   ))
		    ""))

	(t (error "Invalid SIZE" size))))

(defun emchat-v8-number->string-swap (size val)
  "Convert number VAL to string of SIZE."
  (let ((ss (emchat-v8-number->string size val))
	tt)
    (cond ((= size 2)
	   (setq tt (aref ss 0))
	   (aset ss 0 (aref ss 1))
	   (aset ss 1 tt))
	  ((= size 4)
	   (setq tt (aref ss 0))
	   (aset ss 0 (aref ss 3))
	   (aset ss 3 tt)
	   (setq tt (aref ss 1))
	   (aset ss 1 (aref ss 2))
	   (aset ss 2 tt)))
    ss))

(defun emchat-v8-create-message (&rest UNIPREFIX-message-spec)
  "Create emchat message according to SPEC."
  (let (UNIPREFIX-value-spec UNIPREFIX-value)
    (mapconcat (lambda (UNIPREFIX-element)
		 (setq UNIPREFIX-value-spec (eval (aref UNIPREFIX-element 0)))
		 (setq UNIPREFIX-value (eval (aref UNIPREFIX-element 1)))

		 (cond ((numberp UNIPREFIX-value)
			(emchat-v8-number->string UNIPREFIX-value-spec UNIPREFIX-value))
		       ((stringp UNIPREFIX-value)
			(substring  UNIPREFIX-value 0 UNIPREFIX-value-spec))
		       ((null UNIPREFIX-value)
			(make-string UNIPREFIX-value-spec ?\x00))
		       (t (error "Invalid SPEC" UNIPREFIX-value-spec))))
	       UNIPREFIX-message-spec "")
    ))

(defun emchat-v8-string->number-le-flt (str &optional len)
  "Convert STR to number."
  (unless len
    (setq len (length str)))

  (let ((ret 0.0) (i 0))
    (while (< i len)
      (setq ret (+ ret (* (char-int (aref str i)) (expt 256.0 i)))
	    i (1+ i)))
    ret))

(defun emchat-v8-string->number-flt (str &optional len)
  "Convert STR to number."
  (unless len
    (setq len (length str)))

  (let ((ret 0.0) (i 0))
    (while (< i len)
      (setq ret (+ (* 256.0 ret) (char-int (aref str i)))
	    i (1+ i)))
    ret))

(defun emchat-v8-string->number-le-bigz (str &optional len)
  "Convert STR to number."
  ;; optimise for common cases
  (cond ((eq len 4)
	 (+ (* (aref str 3) 16777216)
	    (* (aref str 2) 65536)
	    (* (aref str 1) 256)
	    (aref str 0)))
	((eq len 2)
	 (+ (* (aref str 1) 256)
	    (aref str 0)))
	((eq len 1)
	 (char-to-int (aref str 0)))
	(t
	 (let* ((len (or len (length str)))
		(ret (char-int (aref str 0)))
		(i 1))
	   (while (< i len)
	     (setq ret (+ ret (* (char-int (aref str i))
				 (2^ (* 8 i))))
		   i (1+ i)))
	   ret))))

(defun emchat-v8-string->number-bigz (str &optional len)
  "Convert STR to number."
  ;; optimise for common cases
  (cond ((eq len 4)
	 (+ (* (aref str 0) 16777216)
	    (* (aref str 1) 65536)
	    (* (aref str 2) 256)
	    (aref str 3)))
	((eq len 2)
	 (+ (* (aref str 0) 256)
	    (aref str 1)))
	((eq len 1)
	 (char-to-int (aref str 0)))
	(t
	 (let* ((len (or len (length str)))
		(ret 0)
		(i 0))
	   (while (< i len)
	     (setq ret (+ (lsh ret 8) (char-int (aref str i)))
		   i (1+ i)))
	   ret))))

;; do the right thing at the right time, the wrong otherwise
(cond
 ((featurep 'bigz)
  (defalias #'emchat-v8-string->number #'emchat-v8-string->number-bigz)
  (defalias #'emchat-v8-string->number-le #'emchat-v8-string->number-le-bigz)
  'using-bigz)
 (t
  (defalias #'emchat-v8-string->number #'emchat-v8-string->number-flt)
  (defalias #'emchat-v8-string->number-le #'emchat-v8-string->number-le-flt)
  'using-flt))

(defvar emchat-v8-timeout 15
  "Receive timeout in seconds.")

(defun emchat-v8-grab-bytes (ectx len)
  "Grab LEN bytes from ECTX's incoming buffer."
  (setq len (truncate len))		;for sure

  (let (rstr)
    (while (< (length (emchat-v8-ctx-incoming-buffer ectx)) len)
      (when (null (accept-process-output (emchat-v8-ctx-proc ectx)
					 emchat-v8-timeout))
	(error "EMCHAT: Timeout")))
    (setq rstr (substring (emchat-v8-ctx-incoming-buffer ectx) 0 len))
    (setf (emchat-v8-ctx-incoming-buffer ectx)
	  (substring (emchat-v8-ctx-incoming-buffer ectx) len))
    rstr))

(defun emchat-v8-grab-bytes-1 (vmsg len)
  "Grab LEN bytes from VMSG."
  (setq len (truncate len))

  (let (rstr)
    ;; Check LEN
    (when (> len (length (aref vmsg 0)))
      (error 'invalid-argument "Invalid len" len))

    (setq rstr (substring (aref vmsg 0) 0 len))
    (aset vmsg 0 (substring (aref vmsg 0) len))
    rstr))

;; Generic length storers
(defvar emchat-v8-length-1 nil)
(defvar emchat-v8-length-2 nil)
(defvar emchat-v8-length-3 nil)
(defvar emchat-v8-length-4 nil)

(defun emchat-v8-parse-message (ectx-msg spec)
  "Parse ECTX-MSG according to SPEC.
ECTX-MSG is either EMchat context to grab bytes from, or string.

SPEC is one of
 - Vector         specifies one value.
 - List           specifies multiple values.
 - List of Lists  specifies list of lists of multiple values."
  (when (stringp ectx-msg)
    (setq ectx-msg (vector ectx-msg)))

  (let ((accessor (if (emchat-v8-ctx-p ectx-msg) 'emchat-v8-grab-bytes 'emchat-v8-grab-bytes-1))
	vlen vtype rlist)

    (cond ((vectorp spec)
	   ;; Vector
	   (setq vlen (truncate (eval (aref spec 0))))
	   (setq vtype (aref spec 1))
	   (if (numberp vtype)
	       (when (not (= (emchat-v8-string->number (funcall accessor ectx-msg vlen) vlen)
			     vtype))
		 (error 'invalid-argument "Invalid argument in MSG"))

	     ;; Normal type
	     (cond ((or (eq vtype 'number) (eq vtype 'integer))
		    (setq rlist (emchat-v8-string->number (funcall accessor ectx-msg vlen) vlen)))

		   ((or (eq vtype 'number-swap) (eq vtype 'integer-swap))
		    (setq rlist (emchat-v8-string->number-le (funcall accessor ectx-msg vlen) vlen)))

		   ((eq vtype 'string)
		    (setq rlist (funcall accessor ectx-msg vlen)))

		   ((and (symbolp vtype) (not (keywordp vtype))
			 (member vtype '(emchat-v8-length-1 emchat-v8-length-2 emchat-v8-length-3 emchat-v8-length-4)))
		    ;; If symbol - set it
		    (set vtype (emchat-v8-string->number (funcall accessor ectx-msg vlen) vlen))
		    (setq rlist (symbol-value vtype)))

		   ((listp vtype)
		    (setq rlist (mapcar #'(lambda (not-used)
					    (emchat-v8-parse-message ectx-msg vtype))
					(make-list vlen nil))))

		   ((null vtype) ;; Skip
		    )

		   (t (error 'invalid-argument "Invalid type in SPEC" vtype)))))

	  ((and (listp spec)
		(not (member nil (mapcar (lambda (l) (listp l)) spec))))
	   ;; List of Lists
	   )

	  ((listp spec)
	   ;; List
	   (setq rlist (mapcar (lambda (e) (emchat-v8-parse-message ectx-msg e)) spec)))

	  (t (error 'invalid-argument "Invalid SPEC" spec)))

    rlist))


(put 'emchat-v8-parse-message 'lisp-indent-function 1)

(defun emchat-v8-fetch-data (data len)
  (vector (funcall (if (emchat-v8-ctx-p data)
		       'emchat-v8-grab-bytes
		     'emchat-v8-grab-bytes-1)
		   data len)))

(defun emchat-v8-fetch-string (data &optional len)
  (aref (emchat-v8-fetch-data data (or len (length (aref data 0)))) 0))

(defun emchat-v8-fetch-byte (data)
  (char-to-int (string-to-char (emchat-v8-fetch-string data 1))))
(defun emchat-v8-pack-byte (byte)
  (char-to-string byte))

(defun emchat-v8-fetch-word (data)
  (truncate (emchat-v8-string->number
	     (emchat-v8-fetch-string data 2))))
(defun emchat-v8-pack-word (word)
  (emchat-v8-number->string 2 word))

(defun emchat-v8-fetch-word-le (data)
  (truncate (emchat-v8-string->number-le
	     (emchat-v8-fetch-string data 2))))
(defun emchat-v8-pack-word-le (word)
  (emchat-v8-number->string-swap 2 word))

(defun emchat-v8-fetch-dword (data)
  (emchat-v8-string->number
   (emchat-v8-fetch-string data 4)))
(defun emchat-v8-pack-dword (dword)
  (emchat-v8-number->string 4 dword))

(defun emchat-v8-fetch-time (data)
  (cons (emchat-v8-fetch-word data)
	(emchat-v8-fetch-word data)))

(defun emchat-v8-pack-time (time)
  (concat
   (emchat-v8-pack-word (car time))
   (emchat-v8-pack-word (cdr time))))

(defun emchat-v8-fetch-uin (data)
  (truncate
   (emchat-v8-string->number-le
    (emchat-v8-fetch-string data 4))))

(defun emchat-v8-fetch-uinlist (count uinlist)
  (let (result)
    (dotimes (i count)
      (append (emchat-v8-fetch-uin uinlist) result))))

(defun emchat-v8-pack-uin (uin)
  (emchat-v8-number->string-swap 4 uin))

(defun emchat-v8-fetch-buin (data)
  "Fetch BUIN and return uin number."
  (string-to-number (emchat-v8-fetch-string data (emchat-v8-fetch-byte data))))
(defun emchat-v8-pack-buin (uin)
  (let ((suin (number-to-string uin)))
    (concat (emchat-v8-pack-byte (length suin)) suin)))

(defun emchat-v8-fetch-bstr (data)
  (let* ((len (emchat-v8-fetch-word data))
	 (str (emchat-v8-fetch-string data len)))
    (substring str 0 len)))

(defun emchat-v8-pack-bstr (str)
  (concat (emchat-v8-pack-word (length str)) str))

;; Various packers
(defun emchat-v8-pack-uinlist (uin-list)
  (apply 'concat (mapcar 'emchat-v8-pack-buin uin-list)))

(defun emchat-v8-pack-nts (string)
  (concat string (char-to-string 0)))

(defun emchat-v8-pack-lnts (string)
  (let ((nts (emchat-v8-pack-nts string)))
    (concat
     (emchat-v8-pack-word-le (length nts))
     nts)))

(defun emchat-v8-pack-llnts (string)
  (let ((nts (emchat-v8-pack-nts string)))
    (concat
     (emchat-v8-pack-word-le (+ (length nts) 2))
     (emchat-v8-pack-word-le (length nts))
     nts)))

(defun emchat-v8-fetch-lnts (data)
  (let* ((len (emchat-v8-fetch-word-le data))
	 (str (emchat-v8-fetch-string data len)))
    (substring str 0 (1- len))))

;; FLAP
(defun emchat-v8-fetch-flap (data)
  "Get next queued FLAP from emchat context ECTX."
  (set-process-filter
   (emchat-v8-ctx-proc data) 'emchat-v8-proc-filter-accumulator)

  (unwind-protect
      (progn
	(let ((bb (emchat-v8-fetch-byte data)))
	  (unless (= bb #x2A)
	    (error "Unexpected data while fetching FLAP")))

	(let* ((cin (emchat-v8-fetch-byte data))
	       (seq (emchat-v8-fetch-word data))
	       (flen (emchat-v8-fetch-word data))
	       (fdata (emchat-v8-fetch-data data flen)))
	  (list cin seq fdata)))

    (set-process-filter
     (emchat-v8-ctx-proc data) 'emchat-v8-proc-filter-proccessing)))

(defun emchat-v8-pack-flap (ectx chan-id &rest flap-data)
  "Create FLAP packet of channel id CHAN-ID and DATA."
  (setq flap-data (apply 'concat flap-data))

  (prog1
      (concat
       (emchat-v8-pack-byte #x2A)
       (emchat-v8-pack-byte chan-id)
       (emchat-v8-pack-word (emchat-v8-ctx-sequence ectx))
       (emchat-v8-pack-word (length flap-data))
       flap-data)

    ;; Increase sequence number
    (incf (emchat-v8-ctx-sequence ectx))
    (when (> (emchat-v8-ctx-sequence ectx) 65535)
      (setf (emchat-v8-ctx-sequence ectx)
	    (% (emchat-v8-ctx-sequence ectx) 65536)))))
(put 'emchat-v8-pack-flap 'lisp-indent-function 2)

;; SNAC
(defun emchat-v8-fetch-snac (data)
  "Get next SNAC."
  (let ((family (emchat-v8-fetch-word data))
	(subtype (emchat-v8-fetch-word data))
	(flags (emchat-v8-fetch-word data))
	(rid (emchat-v8-fetch-dword data)))

    ;; If 15th bit is set in FLAGS then DATA contains some additional
    ;; data in format LENGTH [WORD], DATA [LENGTH]
    (unless (zerop (logand flags 32768))
      (let ((len (emchat-v8-fetch-word data)))
	(emchat-v8-fetch-data data len)))

    (list family subtype flags rid)))

(defun emchat-v8-pack-snac (ectx family-subtype &rest snac-data)
  "Create SNAC packet."
  (setq snac-data (apply 'concat snac-data))
  (emchat-v8-pack-flap ectx 2
    (emchat-v8-pack-word (car family-subtype))
    (emchat-v8-pack-word (cdr family-subtype))
    (emchat-v8-pack-word 0)               ; XXX flags
    (emchat-v8-pack-dword 0)              ; XXX rid
    snac-data))
(put 'emchat-v8-pack-snac 'lisp-indent-function 2)

;; TLVs
(defun emchat-v8-fetch-single-tlv (data)
  (let* ((type (emchat-v8-fetch-word data))
	 (len (emchat-v8-fetch-word data))
	 (value (emchat-v8-fetch-data data len)))
    (list type len value)))

(defun emchat-v8-fetch-tlvs (data)
  "From VMSG extract TLV list."
  (let ((tlvs nil))
    (while (> (length (aref data 0)) 0)
      (setq tlvs (cons (emchat-v8-fetch-single-tlv data) tlvs)))
    (nreverse tlvs)))

(defun emchat-v8-fetch-n-tlvs (data n)
  "Read TLVs of total length N and return a list."
  (let ((tlvs nil))
    (while (> n 0)
      (let* ((tlv (emchat-v8-fetch-single-tlv data)))
	(setq n (- n (emchat-v8-tlv-len tlv) 4)
	      tlvs (cons tlv tlvs))))
    (nreverse tlvs)))

(defun emchat-v8-tlv-get (tlv-list type)
  "From TLV list TLV-LIST get tlv of TYPE."
  (while (and tlv-list (not (= (emchat-v8-tlv-type (car tlv-list)) type)))
    (setq tlv-list (cdr tlv-list)))
  (car tlv-list))

(defun emchat-v8-pack-tlv (type &rest values)
  "Create TLV."
  (let ((val (apply 'concat values)))
    (concat
     (emchat-v8-pack-word type)
     (emchat-v8-pack-word (length val))
     val)))
(put 'emchat-v8-pack-tlv 'lisp-indent-function 1)


(defun emchat-v8-login (ectx &optional uin password)
  "In ECTX context send UIN/PASSWORD to ICQ server."
  (emchat-v8-debug "Logging in ..")

  (unless uin
    (setq uin (plist-get (emchat-v8-ctx-userinfo ectx) 'uin)))
  (unless password
    (setq password (plist-get (emchat-v8-ctx-userinfo ectx) 'password)))

  ;; Remove password (for somekind of security)
  (setf (emchat-v8-ctx-userinfo ectx)
	(plist-remprop (emchat-v8-ctx-userinfo ectx) 'password))

  (emchat-v8-send ectx
    (emchat-v8-pack-flap ectx emchat-v8-FLAP-HELLO
      (emchat-v8-pack-dword 1)
      (emchat-v8-pack-tlv 1
	(number-to-string uin))
      (emchat-v8-pack-tlv 2 (emchat-v8-util-encrypt password))
      (emchat-v8-pack-tlv 3 emchat-v8-client-id-string)
      (emchat-v8-pack-tlv #x16
	(emchat-v8-pack-word #x010A))
      (emchat-v8-pack-tlv #x17
	(emchat-v8-pack-word emchat-v8-FLAP-VER-MAJOR))
      (emchat-v8-pack-tlv #x18
	(emchat-v8-pack-word emchat-v8-FLAP-VER-MINOR))
      (emchat-v8-pack-tlv #x19
	(emchat-v8-pack-word emchat-v8-FLAP-VER-LESSER))
      (emchat-v8-pack-tlv #x1A
	(emchat-v8-pack-word emchat-v8-FLAP-VER-BUILD))
      (emchat-v8-pack-tlv #x14
	(emchat-v8-pack-dword emchat-v8-FLAP-VER-SUBBUILD))
      (emchat-v8-pack-tlv #x0F "en")
      (emchat-v8-pack-tlv #x0E "us"))))


(provide 'emchat-v8)

;;; emchat-v8.el ends here
